try:
    from datasets import Dataset
except ImportError:
    raise ImportError("Could not import HuggingFace Datasets: Please install ibm-generative-ai[huggingface] extension.")

from genai.prompt_pattern import PromptPattern
from genai.utils.extensions import register_promptpattern_accessor

__all__ = ["HuggingFaceDatasetExtension"]


@register_promptpattern_accessor("huggingface")
class HuggingFaceDatasetExtension:
    def __init__(self, prompt_pattern: PromptPattern):
        self._obj = prompt_pattern

    def save_dataset(self, list_of_prompts: list, path: str, key: str = "prompt"):
        """
        Given a list of prompts, creates a HuggingFace dataset in a specified path.
        Args:
            list_of_prompts (list): The list of prompts used to create the dataset from.
            path (str): Path in which you want the dataset to be saved.
            key (str, optional): Before the dataset is created, the list is converted to a dictionary.
            This parameter defines what you would like the key in the dictionary to be named.
            Defaults to "prompt".
        """

        list_of_prompts_as_dict = {key: self._obj.list_str(list_of_prompts)}
        ds = Dataset.from_dict(list_of_prompts_as_dict)
        ds.save_to_disk(path)



import os
from pathlib import Path

from dotenv import load_dotenv

from genai.credentials import Credentials
from genai.schemas.tunes_params import (
    CreateTuneHyperParams,
    CreateTuneParams,
    DownloadAssetsParams,
    TunesListParams,
)
from genai.services import FileManager, TuneManager

load_dotenv()
API_KEY = os.getenv("GENAI_KEY", None)
ENDPOINT = os.getenv("GENAI_API", None)

creds = Credentials(api_key=API_KEY, api_endpoint=ENDPOINT)


print("\n=============== UPLOAD LOCAL FILE FOR TUNING ================")
file_path = str(Path(__file__).parent / "file_to_tune.jsonl")
print(" {} \n".format(file_path))

file_uploaded = FileManager.upload_file(credentials=creds, file_path=file_path, purpose="tune")
print("File uploaded: \n", file_uploaded)


print("\n======================== CREATE TUNE ========================")
file_ids = [file_uploaded.id]
hyperparams = CreateTuneHyperParams(verbalizer='classify { "red", "yellow" } Input: {{input}} Output:')

params = CreateTuneParams(
    name="Tune Manager Classification",
    model_id="google/flan-t5-xl",
    method_id="mpt",
    task_id="classification",
    training_file_ids=file_ids,
    parameters=hyperparams,
)

tune_created = TuneManager.create_tune(credentials=creds, params=params)
print("\nNew tune: \n", tune_created)


print("\n======================= GET TUNE BY ID ======================")
t = tune_created.id
tune_get = TuneManager.get_tune(credentials=creds, tune_id=t)
print("\nGet tune result: \n", tune_get)

print("\n======================== LIST TUNES =========================")
list_params = TunesListParams(limit=5, offset=0)
tune_list = TuneManager.list_tunes(credentials=creds, params=list_params)
for t in tune_list.results:
    print("\nTune ID:", t.id, ", Tune Name:", t.name)


print("\n================  LIST EXISTING TUNE METHODS ================")
tune_methods = TuneManager.get_tune_methods(credentials=creds)
print("\n Available Tune Methods:\n")
for t in tune_methods.results:
    print("Method ID:", t.id, ", Method name:", t.name)


print("\n===================  DOWNLOAD TUNE ASSETS ===================")
# content can be: logs or enconder. The download will only be available when the tune is complete.
tune = tune_list.results[0].id
assets_params = DownloadAssetsParams(id=tune, content="encoder")
tune_assets = TuneManager.download_tune_assets(credentials=creds, params=assets_params)
print("\n Tune assets:", tune_assets)


print("\n======================== DELETE TUNE ========================")
tune_delete = TuneManager.delete_tune(credentials=creds, tune_id=tune)
print("\nDelete tune response: \n", tune_delete)



from genai import Credentials
from genai.exceptions import GenAiException
from genai.schemas.responses import (
    WatsonxRenderedPrompts,
    WatsonxTemplate,
    WatsonxTemplatesResponse,
)
from genai.services import ServiceInterface
from genai.utils.errors import to_genai_error


class PromptTemplateManager:
    @staticmethod
    def save_template(template: str, name: str, credentials: Credentials) -> WatsonxTemplate:
        service = ServiceInterface(service_url=credentials.api_endpoint, api_key=credentials.api_key)

        try:
            response = service._prompt_templating.prompt_templates(name, value=template)
            if response.status_code == 200:
                response_result = WatsonxTemplate(**response.json()["results"])
                return response_result
            raise GenAiException(response)
        except Exception as ex:
            raise to_genai_error(ex)

    @staticmethod
    def update_template(credentials: Credentials, id: str, name: str, template: str) -> WatsonxTemplate:
        service = ServiceInterface(service_url=credentials.api_endpoint, api_key=credentials.api_key)

        try:
            response = service._prompt_templating.update_prompt_templates(id=id, name=name, value=template)
            if response.status_code == 200:
                response_result = WatsonxTemplate(**response.json()["results"])
                return response_result
            raise GenAiException(response)
        except Exception as ex:
            raise to_genai_error(ex)

    @staticmethod
    def render_watsonx_prompts(credentials: Credentials, inputs: list = None, data: dict = {}) -> list[str]:
        service = ServiceInterface(service_url=credentials.api_endpoint, api_key=credentials.api_key)

        try:
            response = service._prompt_templating.prompt_output(inputs, template=data)
            if response.status_code == 200:
                response_result = WatsonxRenderedPrompts(**response.json())
                return response_result.results
            raise GenAiException(response)
        except Exception as ex:
            raise to_genai_error(ex)

    @staticmethod
    def load_template(credentials: Credentials, id: str = None, name: str = None) -> WatsonxTemplate:
        if id:
            return PromptTemplateManager.load_template_by_id(credentials=credentials, id=id)
        if name:
            return PromptTemplateManager.load_template_by_name(credentials=credentials, name=name)
        else:
            raise GenAiException(
                "Provide either name or id of prompt to be fetch."
                + "\nIf you want to list all templates, use the load_all_templates method."
            )

    @staticmethod
    def load_all_templates(credentials: Credentials) -> WatsonxTemplatesResponse:
        service = ServiceInterface(service_url=credentials.api_endpoint, api_key=credentials.api_key)

        try:
            response = service._prompt_templating.get_prompt_templates()
            if response.status_code == 200:
                return WatsonxTemplatesResponse(**response.json())
            raise GenAiException(response)
        except Exception as ex:
            raise to_genai_error(ex)

    @staticmethod
    def load_template_by_id(credentials: Credentials, id: str) -> WatsonxTemplate:
        service = ServiceInterface(service_url=credentials.api_endpoint, api_key=credentials.api_key)

        try:
            response = service._prompt_templating.get_prompt_templates(id)
            if response.status_code == 200:
                return WatsonxTemplate(**response.json()["results"])
            raise GenAiException(response)
        except Exception as ex:
            raise to_genai_error(ex)

    @staticmethod
    def load_template_by_name(credentials: Credentials, name: str) -> WatsonxTemplate:
        try:
            saved_templates = PromptTemplateManager.load_all_templates(credentials=credentials)
            template = list(filter(lambda tp: tp.name == name, saved_templates.results))

            if len(template) == 1:
                return template[0]

            if len(template) == 0:
                raise Exception(f"No template found with name {name}")
            raise Exception(f"More than one template found with name {name}")
        except Exception as ex:
            raise to_genai_error(ex)

    @staticmethod
    def delete_template(credentials: Credentials, id: str = None, name: str = None) -> str:
        if id:
            return PromptTemplateManager.delete_template_by_id(credentials=credentials, id=id)
        elif name:
            return PromptTemplateManager.delete_template_by_name(credentials=credentials, name=name)
        else:
            raise GenAiException("Provide either name or id of prompt to be deleted.")

    @staticmethod
    def delete_template_by_id(credentials: Credentials, id: str) -> str:
        service = ServiceInterface(service_url=credentials.api_endpoint, api_key=credentials.api_key)

        try:
            response = service._prompt_templating.delete_prompt_templates(id)
            if response.status_code == 204:
                return id
            raise GenAiException(response)
        except GenAiException as ex:
            raise Exception(ex)

    @staticmethod
    def delete_template_by_name(credentials: Credentials, name: str) -> str:
        try:
            saved_templates = PromptTemplateManager.load_all_templates(credentials=credentials)
            template = list(filter(lambda tp: tp.name == name, saved_templates.results))

            if len(template) == 1:
                return PromptTemplateManager.delete_template_by_id(credentials=credentials, id=template[0].id)
            if len(template) == 0:
                raise Exception(f"No template found for name {name}")
            raise Exception(f"More than one template found for name {name}")
        except GenAiException as e:
            raise Exception(e)



import asyncio
import logging
import os
import random

from dotenv import load_dotenv

from genai.model import Credentials, Model
from genai.schemas import GenerateParams, TokenParams
from genai.services.async_generator import AsyncResponseGenerator
from genai.utils.errors import to_genai_error

num_requests = 0


class FlakyAsyncResponseGenerator(AsyncResponseGenerator):
    async def _get_response_json(self, model, inputs, params, options):
        try:
            global num_requests
            num_requests += 1
            if num_requests % 2 == 0:
                await asyncio.sleep(random.randint(0, 5))
                response_raw = await self._service_fn(model, inputs, params, options)
            else:
                await asyncio.sleep(random.randint(0, 5))
                response_raw = None  # bad response
            response = response_raw.json()
        except Exception as _:  # noqa
            response = None
        return response


class FlakyModel(Model):
    def generate_async(
        self,
        prompts,
        ordered: bool = False,
        callback=None,
        options=None,
    ):
        try:
            with FlakyAsyncResponseGenerator(
                self.model,
                prompts,
                self.params,
                self.service,
                ordered=ordered,
                callback=callback,
                options=options,
            ) as asynchelper:
                for response in asynchelper.generate_response():
                    yield response
        except Exception as ex:
            raise to_genai_error(ex)

    def tokenize_async(self, prompts, ordered=False, callback=None, options=None):
        try:
            with FlakyAsyncResponseGenerator(
                self.model,
                prompts,
                self.params,
                self.service,
                fn="tokenize",
                ordered=ordered,
                callback=callback,
                options=options,
            ) as asynchelper:
                for response in asynchelper.generate_response():
                    yield response
        except Exception as ex:
            raise to_genai_error(ex)


logging.getLogger("genai").setLevel(logging.INFO)


# make sure you have a .env file under genai root with
# GENAI_KEY=<your-genai-key>
load_dotenv()
api_key = os.getenv("GENAI_KEY", None)
creds = Credentials(api_key=api_key)  # credentials object to access GENAI


# Instantiate parameters for text generation
generate_params = GenerateParams(decoding_method="sample", max_new_tokens=5, min_new_tokens=1)
tokenize_params = TokenParams(return_tokens=True)


flan_ul2 = FlakyModel("google/flan-ul2", params=generate_params, credentials=creds)
prompts = ["Generate a random number > {}: ".format(i) for i in range(17)]
print("======== Async Generate with ordered=True ======== ")
counter = 0
for response in flan_ul2.generate_async(prompts, ordered=True):
    counter += 1
    if response is not None:
        print(counter, ":", response.input_text, " --> ", response.generated_text)
    else:
        print(counter, ":", None)

num_requests = 0

# Instantiate a model proxy object to send your requests
flan_ul2 = FlakyModel("google/flan-ul2", params=tokenize_params, credentials=creds)
prompts = ["Generate a random number > {}: ".format(i) for i in range(23)]
print("======== Async Tokenize with ordered=True ======== ")
counter = 0
for response in flan_ul2.tokenize_async(prompts, ordered=True):
    counter += 1
    if response is not None:
        print(counter, ":", response.input_text, " --> ", response.tokens)
    else:
        print(counter, ":", None)



import pytest

from genai import Metadata


@pytest.mark.integration
class TestTermsOfUseIntegration:
    def test_accept_tou(self, credentials):
        metadata = Metadata(credentials)

        tou_response = metadata.accept_terms_of_use()
        assert tou_response.results.tou_accepted is True



import os
from pathlib import Path

from dotenv import load_dotenv

from genai.credentials import Credentials
from genai.schemas.tunes_params import CreateTuneParams
from genai.services import FileManager, TuneManager

load_dotenv()
API_KEY = os.getenv("GENAI_KEY", None)
ENDPOINT = os.getenv("GENAI_API", None)

creds = Credentials(api_key=API_KEY, api_endpoint=ENDPOINT)

# # UPLOAD FILE
# file_path = "<path-to-file>"
print("======================== UPLOAD LOCAL FILE ========================")
file_path = str(Path(__file__).parent / "file_to_tune.jsonl")
print(" {} \n".format(file_path))

upload_file = FileManager.upload_file(credentials=creds, file_path=file_path, purpose="tune")
print("File uploaded: \n", upload_file)

print("\n===================== GET UPLOADED FILE ID ========================")
file_list = FileManager.list_files(credentials=creds)
for f in file_list.results:
    if f.file_name == "file_to_tune.jsonl":
        file_id = f.id
        break
print("Uploaded file has id = {}\n".format(file_id))

print("\n================= CREATE TUNE FOR GENERATION TASK ==================")

tunes_params = CreateTuneParams(
    name="flan-t5-xl",
    model_id="google/flan-t5-xl",
    method_id="pt",
    task_id="generation",
    training_file_ids=[file_id],
)

tune_create = TuneManager.create_tune(credentials=creds, params=tunes_params)
print("Created tune has the id: \n ", tune_create.id)



import logging
import pathlib
import runpy

import pytest

logger = logging.getLogger(__name__)

all_scripts = pathlib.Path(__file__, "..", "..", "..", "examples", "user").resolve().glob("*.py")
scripts = (l for l in all_scripts if "async_" not in l.name)


def idfn(val):
    script_path = pathlib.Path(val)
    return script_path.name


@pytest.mark.integration
@pytest.mark.parametrize("script", scripts, ids=idfn)
def test_example_execution(script):
    logger.info(f"Executing Example script: {script}")
    runpy.run_path(str(script))



import os
import pathlib
from unittest.mock import MagicMock, mock_open, patch

import pytest

from genai import Credentials
from genai.exceptions import GenAiException
from genai.routers.tunes import TunesRouter
from genai.schemas.responses import (
    TuneInfoResult,
    TuneMethodsGetResponse,
    TunesListResponse,
)
from genai.schemas.tunes_params import (
    CreateTuneParams,
    DownloadAssetsParams,
    TunesListParams,
)
from genai.services.tune_manager import TuneManager
from tests.assets.response_helper import SimpleResponse


@pytest.mark.unit
class TestTunes:
    def setup_method(self):
        self.service_router = TunesRouter(service_url="SERVICE_URL", api_key="API_KEY")
        self.path = pathlib.Path(__file__).parent.resolve()
        self.asset_path = pathlib.Path(__file__, "..", "assets").resolve()

    @pytest.fixture
    def list_params(self):
        return TunesListParams(limit=10, offset=0)

    @pytest.fixture
    def download_assets_params(self):
        return DownloadAssetsParams(id="some-tune-id", content="encoder")

    @pytest.fixture
    def create_params(self):
        return CreateTuneParams(
            name="Test",
            model_id="some-model-id",
            method_id="some-method-id",
            task_id="generation",
            training_file_ids=["some-file-id"],
        )

    @pytest.fixture
    def credentials(self):
        return Credentials("GENAI_KEY")

    # Test list tunes function
    @patch("genai.services.RequestHandler.get")
    def test_list_tunes(self, mock_requests, credentials, list_params):
        list_response = SimpleResponse.tunes(params=list_params)
        expected_response = TunesListResponse(**list_response)

        mock_response = MagicMock(status_code=200)
        mock_response.json.return_value = list_response
        mock_requests.return_value = mock_response

        f = TuneManager.list_tunes(credentials=credentials, params=list_params)
        assert f == expected_response

    @patch("genai.services.RequestHandler.get")
    def test_tune_list_api_call(self, mock_requests, list_params):
        mock_response = MagicMock(status_code=200)
        mock_response.json.return_value = SimpleResponse.tunes()
        mock_requests.return_value = mock_response

        g = self.service_router.list_tunes(params=list_params)
        assert g == mock_response

    def test_list_tunes_wrong_params(self):
        with pytest.raises(GenAiException) as e:
            self.service_router.list_tunes(params="params")
            assert e.message == "params must be of type TunesListParams."

    def test_list_tunes_wrong_params_type(self):
        with pytest.raises(GenAiException) as e:
            self.service_router.list_tunes(params=123)
            assert e.message == "params must be of type TunesListParams."

    # Test delete tunes function
    @patch("genai.services.RequestHandler.delete")
    def test_delete_tune(self, mocker, credentials):
        expected_response = {"status": "success"}
        response = MagicMock(status_code=204)
        response.json.return_value = expected_response
        mocker.return_value = response

        with pytest.raises(Exception) as e:
            file_response = TuneManager.delete_tune(credentials=credentials, tune_id="tune_id")
            assert e.message == "Tune can not be deleted until is completed or failed."
            assert file_response == expected_response

    @patch("genai.services.RequestHandler.delete")
    def test_delete_tune_api_call(self, mocker):
        response = MagicMock(status_code=204)
        response.json.return_value = {"status": "success"}
        mocker.return_value = response

        tune_response = self.service_router.delete_tune(tune_id="tune_id")

        assert tune_response == response

    def test_tune_delete_wrong_params(self):
        with pytest.raises(Exception) as e:
            self.service_router.delete_tune(tune_id=123)
            assert e.message == "Tune not found, tune_id must be of type str"

    # Test get tune function

    @patch("genai.services.RequestHandler.get")
    def test_get_tune(self, mock_requests, credentials):
        get_tune_response = SimpleResponse.tunes(tune_id="tune_id")
        expected_response = TuneInfoResult(**get_tune_response["results"])

        mock_response = MagicMock(status_code=200)
        mock_response.json.return_value = get_tune_response
        mock_requests.return_value = mock_response

        tune_response = TuneManager.get_tune(credentials=credentials, tune_id="tune_id")
        assert tune_response == expected_response

    @patch("genai.services.RequestHandler.get")
    def test_get_tune_api_call(self, mock_requests):
        mock_response = MagicMock(status_code=200)
        mock_response.json.return_value = SimpleResponse.tunes()
        mock_requests.return_value = mock_response
        g = self.service_router.get_tune(tune_id="tune_id")
        assert g == mock_response

    def test_get_tune_wrong_params(self):
        with pytest.raises(Exception) as e:
            self.service_router.get_tune(tune_id=123)
            assert e.message == "Tune not found, tune_id must be of type str"

    # Test create tune function

    @patch("genai.services.RequestHandler.post")
    def test_create_tune(self, mocker, create_params, credentials):
        create_response = SimpleResponse.tunes(params=create_params)
        expected_create_response = TuneInfoResult(**create_response["results"])

        response = MagicMock(status_code=200)
        response.json.return_value = create_response
        mocker.return_value = response

        tune_response = TuneManager.create_tune(credentials=credentials, params=create_params)

        assert tune_response == expected_create_response

    @patch("genai.services.RequestHandler.post")
    def test_create_tune_api_call(self, mock_requests, create_params):
        mock_response = MagicMock(status_code=200)
        mock_response.json.return_value = SimpleResponse.tunes()
        mock_requests.return_value = mock_response

        g = self.service_router.create_tune(params=create_params)
        assert g == mock_response

    def test_create_tune_wrong_params(self):
        with pytest.raises(GenAiException) as e:
            self.service_router.create_tune(params="params")
            assert e.message == "params must be of type dict as CreateTuneParams."

    def test_create_tune_wrong_params_type(self):
        with pytest.raises(GenAiException) as e:
            self.service_router.create_tune(params=123)
            assert e.message == "params must be of type dict as CreateTuneParams."

    # Test get tune methods function

    @patch("genai.services.RequestHandler.get")
    def test_get_tune_methods(self, mock_requests, credentials):
        tune_methods_response = SimpleResponse.get_tune_methods()
        expected_response = TuneMethodsGetResponse(**tune_methods_response)

        mock_response = MagicMock(status_code=200)
        mock_response.json.return_value = tune_methods_response
        mock_requests.return_value = mock_response

        tune_response = TuneManager.get_tune_methods(credentials=credentials)
        assert tune_response.results == expected_response.results

    @patch("genai.services.RequestHandler.get")
    def test_get_tune_methods_api_call(self, mock_requests):
        mock_response = MagicMock(status_code=200)
        mock_response.json.return_value = SimpleResponse.get_tune_methods()
        mock_requests.return_value = mock_response
        g = self.service_router.get_tune_methods()
        assert g == mock_response

    # Test download tune assets function

    @patch("genai.services.RequestHandler.get")
    @patch("builtins.open", new=mock_open(read_data="some data"), create=True)
    def test_download_assets(self, mock_request, credentials, download_assets_params, tmp_path):
        output_path = pathlib.Path(self.asset_path, tmp_path, "tune_assets").resolve()
        mock_request.json.return_value = MagicMock(status_code=200)

        with pytest.raises(Exception) as e:
            TuneManager.download_tune_assets(
                credentials=credentials, params=download_assets_params, output_path=output_path
            )
            assert e.message == "Tune can not be deleted until is completed or failed."
            assert os.path.exists(output_path)
            assert os.path.isdir(output_path)



import logging
import os

from dotenv import load_dotenv

from genai.credentials import Credentials
from genai.model import Model
from genai.schemas import GenerateParams, TokenParams

logging.getLogger("genai").setLevel(logging.INFO)


# make sure you have a .env file under genai root with
# GENAI_KEY=<your-genai-key>
# GENAI_API=<genai-api-endpoint>
load_dotenv()
api_key = os.getenv("GENAI_KEY", None)
api_endpoint = os.getenv("GENAI_API", None)
creds = Credentials(api_key=api_key, api_endpoint=api_endpoint)  # credentials object to access GENAI


# Instantiate parameters for text generation
generate_params = GenerateParams(decoding_method="sample", max_new_tokens=500, min_new_tokens=10)
tokenize_params = TokenParams(return_tokens=True)


flan_ul2 = Model("google/flan-ul2", params=generate_params, credentials=creds)
prompts = ["Explain life in one sentence:", "Write a python function to permute an array."] * 5
print("======== Async Generate (responses need not be in order) ======== ")
counter = 0
for response in flan_ul2.generate_async(prompts):
    counter += 1
    if response is not None:
        print(counter, response.input_text, " --> ", response.generated_text)
    else:
        print(counter, ":", None)

print("======== Async Generate with ordered=True (responses in order) ======== ")
counter = 0
for response in flan_ul2.generate_async(prompts, ordered=True):
    counter += 1
    if response is not None:
        print(
            counter,
            ": Input Prompt ==> ",
            prompts[counter - 1],
            "\n(Input text, Generated text) ==>",
            response.input_text,
            " --> ",
            response.generated_text,
        )
    else:
        print(counter, ":", None)


# Instantiate a model proxy object to send your requests
flan_ul2 = Model("google/flan-ul2", params=tokenize_params, credentials=creds)
prompts = ["Explain life in one sentence:", "Write a python function to permute an array." * 5] * 5
print("======== Async Tokenize (responses need not be in order) ======== ")
counter = 0
for response in flan_ul2.tokenize_async(prompts):
    counter += 1
    if response is not None:
        print(counter, ":", response.input_text, " --> ", response.tokens)
    else:
        print(counter, ":", None)

print("======== Async Tokenize with ordered=True (responses in order) ======== ")
counter = 0
for response in flan_ul2.tokenize_async(prompts, ordered=True):
    counter += 1
    if response is not None:
        print(
            counter,
            ": Input Prompt ==>",
            prompts[counter - 1],
            "\n(Input text, Tokens) ==>",
            response.input_text,
            " --> ",
            response.tokens,
        )
    else:
        print(counter, ":", None)



import logging
from typing import Union

from httpx import Response
from pydantic import ValidationError

from genai.schemas.responses import ErrorResponse

logger = logging.getLogger(__name__)


class GenAiException(Exception):
    def __init__(self, error: Union[Exception, Response, str]) -> None:
        if isinstance(error, Response):
            try:
                self.error = ErrorResponse(**error.json())
                self.error_message = self.error.message
            except ValidationError:
                self.error = error
                self.error_message = str(error.content)
        else:
            self.error = error
            self.error_message = str(error)
        if "TOU_NOT_ACCEPTED" in self.error_message:
            split_message = self.error_message.split("Terms of use not accepted")
            self.error_message = "".join(
                [
                    split_message[0],
                    "Terms of use not accepted. Please accept the terms of use in a browser.",
                    split_message[1],
                ]
            )
        logger.error(self.error_message)
        super().__init__(self.error_message)



from typing import Literal, Optional

from pydantic import BaseModel, Extra, Field

from genai.schemas import Descriptions as tx

# API Reference : https://workbench.res.ibm.com/docs


class HistoryParams(BaseModel):
    class Config:
        anystr_strip_whitespace = True
        extra = Extra.forbid

    limit: Optional[int] = Field(None, description=tx.LIMIT, le=100)
    offset: Optional[int] = Field(None, description=tx.OFFSET)
    status: Optional[Literal["SUCCESS", "ERROR"]] = Field(None, description=tx.STATUS)
    origin: Optional[Literal["API", "UI"]] = Field(None, description=tx.ORIGIN)



"""Wrapper around IBM GENAI APIs for use in langchain"""
import logging
import re

try:
    from langchain.prompts import PromptTemplate
except ImportError:
    raise ImportError("Could not import langchain: Please install ibm-generative-ai[langchain] extension.")

from genai.prompt_pattern import PromptPattern
from genai.utils.extensions import register_promptpattern_accessor

logger = logging.getLogger(__name__)

__all__ = ["PromptExtension"]


@register_promptpattern_accessor("langchain")
class PromptExtension:
    def __init__(self, prompt_pattern: PromptPattern):
        self._obj = prompt_pattern

    def as_template(self):
        """Convert genai PromptPattern to langchain PromptTemplate.
        Returns:
            PromptTemplate: An instance of langchain PromptTemplate.
        """
        s = self._obj.dump
        vars = list(self._obj.find_vars())
        for var in vars:
            s = re.sub(r"{{\s*" + var + r"\s*}}", r"{" + var + r"}", s)
        return PromptTemplate(input_variables=vars, template=s)

    @staticmethod
    def from_template(template: PromptTemplate):
        """Convert langchain PromptTemplate to genai PromptPatterm.
        Args:
            template (PromptTemplate): An instance of langchain PromptTemplate.
        Returns:
            PromptPattern: An instance of genai prompt pattern.
        """
        s = template.template
        for var in template.input_variables:
            s = re.sub("{" + var + "}", "{{" + var + "}}", s)
        return PromptPattern.from_str(s)



import copy
import csv
import random
import re
from typing import List, Literal, Union

import requests
from yaml import CLoader as Loader
from yaml import load

from genai import Credentials
from genai.exceptions import GenAiException
from genai.schemas.responses import WatsonxTemplate
from genai.services import PromptTemplateManager
from genai.utils.json_utils import json_extract, json_get_all_keys, json_load


class PromptPattern:
    _accessors = set()

    def _create(self, *args, **kwargs):
        self.url: str = None
        self.literal: bool = False
        self.dump = None
        self.yaml = None
        self.token: str = None

        self.watsonx: WatsonxTemplate = None
        self.credentials: Credentials = None

        for key, value in kwargs.items():
            setattr(self, key, value)

        self.raw_content = self._fetch_prompt()

    @classmethod
    def from_url(cls, url: str, token=""):
        """
        Create a promptPattern from a resource located at a remote url.

        Args:
            url (str): Location of the resource.
            token (str, optional): User token to access the resource. Defaults to "".

        Returns:
            PromptPattern: Returns the created PromptPattern instance.
        """

        pt = PromptPattern()
        pt._create(url=url, token=token)
        return pt

    @classmethod
    def from_file(cls, path: str):
        """
        Create a promptPattern from a file.

        Args:
            path (str): Path to file.

        Returns:
            PromptPattern: Returns the created PromptPattern instance.
        """

        pt = PromptPattern()
        pt._create(url=path)
        return pt

    @classmethod
    def from_str(cls, prompt: str):
        """
        Create a promptPattern from a string.

        Args:
            prompt (str): The prompt template as string.

        Returns:
            PromptPattern: Returns the created PromptPattern instance.
        """
        pt = PromptPattern()
        pt._create(url=prompt, literal=True)
        return pt

    @classmethod
    def from_watsonx(
        cls,
        credentials: Credentials,
        name: str = None,
        template: str = None,
        id: str = None,
    ):
        # Cases :
        # fetching an existing template : name OR id
        # updating an existing template : template + (name OR id)
        # creating a new template       : template + name

        if not template:
            wx = PromptTemplateManager.load_template(credentials=credentials, id=id, name=name)
            pt = PromptPattern()
            pt._create(url=wx.value, watsonx=wx, credentials=credentials, literal=True)
            return pt

        try:
            saved_template = PromptTemplateManager.load_template(credentials=credentials, id=id, name=name)
            id = saved_template.id
            name = name if name else saved_template.name
            wx = PromptTemplateManager.update_template(credentials=credentials, id=id, name=name, template=template)
        except Exception:
            # Template with name doesn't exist. Save template
            wx = PromptTemplateManager.save_template(template=template, name=name, credentials=credentials)

        pt = PromptPattern()
        pt._create(url=wx.value, watsonx=wx, credentials=credentials, literal=True)
        return pt

    def _fetch_prompt(self):
        """
        Gets the raw content as a string at the path indicated by the url.
        The url can point to any http endpoint or a local path

        Returns:
            str: Contents of file located at url as a str
        """

        if not self.literal:
            if self.url.startswith("http"):
                if self.token is not None:
                    headers = {"Authorization": f"Bearer {self.token}"}
                    resp = requests.get(self.url, headers=headers)
                else:
                    resp = requests.get(self.url)
                raw_content = resp.content
                resp.close()
            else:
                with open(self.url, "r", encoding="utf-8") as file:
                    raw_content = file.read()

            self.yaml = load(raw_content, Loader=Loader)
            if "content" not in self.yaml:
                raise Exception("Invalid Prompt YAML. It must include 'content' fields.")
            self.dump = self.yaml["content"]
        else:
            raw_content = self.url
            self.dump = raw_content

        return raw_content

    def reset(self):
        if self.literal:
            self.dump = self.raw_content
        else:
            self.yaml = load(self.raw_content, Loader=Loader)
            self.dump = self.yaml["content"]

    def refetch(self):
        self.raw_content = self._fetch_prompt()

    def validate(self):
        """
        Validate if the yaml at apiVersion is correct.
        Returns:
            returns True if passes and raises Exception otherwise
        """

        # we do not validate content of literal, so we pass as true
        if self.literal:
            return True
        if self.yaml:
            if "apiVersion" not in self.yaml:
                raise Exception("Invalid Prompt YAML. It must include 'apiVersion'.")
            if self.yaml["apiVersion"].lower() == "v0" and "content" not in self.yaml:
                raise Exception("Invalid Prompt YAML. It must include 'content' fields.")

        return True

    def find_vars(self):
        """
        Find all variables in prompt template and return in a list.

        Returns:
            list of variables in prompt template
        """

        pattern = re.compile(r"{{\s*" + "[^{}]*" + r"\s*}}")
        prompt = self.dump
        vars = re.findall(pattern, prompt)
        # remove prefix and suffix
        cleaned = []
        for var in vars:
            name = var[2:-2].strip()
            cleaned.append(name)
        return set(cleaned)

    def sub(self, var: str, value: str) -> "PromptPattern":
        """
        Substitute all variables named 'var' (indicated as {{var}} in prompt contents)

        Args:
            var: name of variable ( cannot contain the characters '{' or '}' )
            value: value to replace variable with

        Returns:
            str: contents of prompt with substitution
        """

        pattern = re.compile(r"{{\s*" + var + r"\s*}}")
        prompt = self.dump
        self.dump = re.sub(pattern, value.replace("\\", r"\\"), prompt)
        return self

    def _get_idx(self, strategy, start_index, idx, max_length, random_idx_list):
        """
        Return an index based on the given strategy

        Args:
            strategy (string): Either 'sequential', 'sample' or 'random'
            start_index (int): Start index
            idx (int): Index of element
            max_length (int): Upper bound of the return value
            random_idx_list (list[int]): List of random indexes

        Returns:
            int: The desired index
        """

        if strategy == "sequential":
            index = min(start_index + idx, max_length)
        elif strategy == "random":
            index = random.randint(0, max_length)
        elif strategy == "sample":
            index = min(start_index + idx, len(random_idx_list) - 1)
            index = random_idx_list[index]
        return index

    def sub_all_from_json(
        self,
        json_path: str,
        key_to_var: Union[dict, Literal["infer"]] = "infer",
        encoding: str = "utf-8-sig",
    ) -> list["PromptPattern"]:
        """
        Sequentially builds a list of PromptPattern from a json file. Iterates through the whole file
        from start to end.

        Args:
            json_path (str): Path to json file.
            key_to_var (dict, str): "infer" string or dict mapping of key value to list of variables.
                If "infer", the key_to_var mapping is inferred based on the template's variables' name.
                For example, assuming a variable "name" and a key "name" in the passed json file, "infer"
                is equivalent to key_to_var={"name":["name"]}. Defaults to "infer".
            encoding (str, Optional): File encoding. Defaults to "utf-8-sig".

        Returns:
            List[PromptPattern] : List of PromptPattern
        """

        start_index = 0
        n = -1
        return self.sub_from_json(json_path, key_to_var, "sequential", start_index, n, encoding)

    def sub_from_json(
        self,
        json_path: str,
        key_to_var: Union[dict, Literal["infer"]] = "infer",
        strategy: Literal["random", "sequential", "sample"] = "sequential",
        start_index: int = 0,
        n: int = 1,
        encoding: str = "utf-8-sig",
    ) -> Union["PromptPattern", list["PromptPattern"]]:
        """
        This will replace every variable from the key_to_var dict using data from the provided file.

        Args:
            json_path (str): Path to json or jsonl file
            key_to_var (Union[dict, Literal["infer"]]): "infer" string or dict mapping keys to variables.
                If "infer", the key_to_var mapping is inferred based on the template's variables' name.
                For example, assuming a variable "name" and a key "name" in the json file, "infer"
                is equivalent to key_to_var={"name":["name"]}. Defaults to "infer". In its most
                general form, a user can specify a mapping like key_to_var={"key": list_of_vars,. ..}
                where the list against each key is a list of variables which can be substituted from
                that key.
            strategy (Literal[sequential, sample, random], optional): Define how the template will be populated from
                the given json.
                Sequential  : sequentially subs the template's variables from the json, starting at a given start_index.
                Sample      : randomly samples rows and instantiate prompts. Without repetition of the data.
                Random      : for each prompt variable, substitute the value from the corresponding key at random
                Defaults to "sequential".
            start_index (int, optional): Only used with strategy sequential. State the first row to be used for the
                substitution in the template. Defaults to 0.
            n (int, Optional): Desired number of prompts to be returned.
                - n = 1: default, returns a PromptPattern
                - n > 1: returns a list of PromptPattern up to min(n, maximum producible prompts from file)
                - n = -1: goes through the entire file.
                If n is -1 and strategy is Random, a GenAiException is raised.
            encoding (str, Optional): File encoding. Defaults to "utf-8-sig".

        Returns:
            PromptPattern | list[PromptPattern]: A populated PromptPattern or a list of PromptPatterns
        """

        if strategy == "random" and n <= 0:
            raise GenAiException("n must be at least 1")

        with open(json_path, "r", encoding=encoding) as file:
            data = json_load(file)

            if len(data) < 1:
                raise ValueError("JSON file seems to be empty: {}".format(json_path))
            if key_to_var == "infer":
                key_to_var = self._json_infer_mode_helper(data[0])

            random_row_idx = [*range(len(data) - 1)]
            random.shuffle(random_row_idx)

            if strategy == "sample":
                start_index = 0

            PromptPattern.validate_start_index(strategy, start_index, data)

            done = False
            complete_pt = []

            while not done:
                pt = copy.copy(self)

                for key in key_to_var.keys():
                    variables = key_to_var[key]

                    for idx, val in enumerate(variables):
                        top = len(data) - 1
                        index = pt._get_idx(strategy, start_index, idx, top, random_row_idx)
                        row = data[index]
                        pt.sub(val, json_extract(row, key, join=True))

                    current_row_idx = start_index + idx
                    if (strategy == "sequential" or strategy == "sample") and current_row_idx >= len(data):
                        done = True
                        break

                if not done:
                    complete_pt.append(pt)

                if n > -1 and len(complete_pt) >= n:
                    done = True

                start_index += len(variables)

        return self._return_single_prompt_or_completed_list(complete_pt, n)

    def sub_all_from_csv(
        self,
        csv_path: str,
        col_to_var: Union[dict, Literal["infer"]] = "infer",
        headers: bool = True,
        encoding: str = "utf-8-sig",
        delimiter: str = ",",
    ) -> list["PromptPattern"]:
        """
        Sequentially builds a list of PromptPattern from a csv file. Iterates through the whole file
        from start to end.

        Args:
            csv_path (str): Path to csv file.
            col_to_var (Union[dict, Literal["infer"]]): "infer" string or dict of column name to list of variables.
                If "infer", the col_to_var mapping is inferred based on the template's variables' name.
                For example, assuming a variable "name" and a column "name" in the passed csv file, "infer"
                is equivalent to col_to_var={"name":["name"]}. Defaults to "infer". In its most
                general form, a user can specify a mapping like col_to_var={"column": list_of_vars,. ..}
                where the list against each column is a list of variables which can be substituted from
                that column.
            headers (bool, Optional): Indicates if the csv contains a header row. Only used if inferring col_to_var.
                Defalts to True.
            encoding (str, Optional): File encoding. Defaults to "utf-8-sig".
            delimiter (str, Optional): Delimiter for the csv. Defaults to ",".

        Returns:
            List[PromptPattern] : List of PromptPattern
        """

        start_index = 0
        n = -1
        return self.sub_from_csv(csv_path, col_to_var, headers, "sequential", start_index, n, encoding, delimiter)

    def _json_infer_mode_helper(self, first_dict):
        vars = self.find_vars()
        if all([v.isdigit() for v in vars]):
            raise GenAiException("Variable inference with digit is not supported for JSON document.")

        keys = json_get_all_keys(first_dict)
        overlap = set(keys).intersection(vars)
        key_to_var = {v: [v] for v in overlap}
        return key_to_var

    def _tabular_infer_mode_helper(self, first_row):
        vars = self.find_vars()
        if all([v.isdigit() for v in vars]):
            columns = [str(i) for i in range(len(first_row))]
            overlap = set(columns).intersection(vars)
        else:
            columns = first_row[:]
            overlap = set(first_row).intersection(vars)

        col_to_var = {v: [v] for v in overlap}
        if col_to_var == {}:
            raise GenAiException(
                """
                Could not match variables {} to data keys {}.
                If you are inferring and using a headless csv, make sure the variables
                correspond to the index of the desired csv column.
                """.format(
                    vars, first_row
                )
            )

        return col_to_var, columns

    def _return_single_prompt_or_completed_list(self, prompts: list, n: int):
        """
        Return a single PromptTemplate is the requested number is one, changing the list of prompts as a single prompt.
        If the number of prompts is zero, raise an exception.

        Args:
            prompts (list): List of prompts to be checked.
            n (int): Number of prompts to be generated.

        Returns:
            self: If the prompt is complete.
            Raises GenAiException: If the given parameters coud not produce a prompt.
        """
        if n == 1:
            if len(prompts) == 0:
                raise GenAiException("Could not produce a prompt from given parameters.")
            self.__dict__ = copy.deepcopy(prompts[0].__dict__)
            return self
        else:
            return prompts

    def _random_row_idx_helper(self, data: list[list[str]]) -> list:
        """
        Returns a random row from indexes the given data.

        Args:
            data (list[list[str]]): List of rows.

        Returns:
            list[str]: A random row.
        """
        random_row_idx = [*range(len(data) - 1)]
        random.shuffle(random_row_idx)
        return random_row_idx

    def _sub_from_tabular_data(
        self,
        data: list[list[str]],
        columns: list,
        col_to_var: dict,
        start_index: int = 0,
        n: int = 1,
        strategy: Literal["random", "sequential", "sample"] = "sequential",
    ) -> Union["PromptPattern", list["PromptPattern"]]:
        """
        Substitutes variables from a tabular data.
        Returns a complete prompt if n=1, else returns a list of prompts.

        Args:
            data (list[list[str]]): List of rows.
            columns (list): List of column names from the tabular data.
            col_to_var (dict): A dictionary mapping column names to a list of variables to substitute.
            start_index (int, Optional): Index of the row to start from. Defaults to 0.
            n (int, Optional): Number of prompts to generate. Defaults to 1.
            strategy (Literal["random", "sequential", "sample"], Optional): Strategy to use for generating prompts.
                Defaults to "sequential".

        Returns:
            list[PromptPattern]: List of PromptPattern
        """

        done = False
        complete_prompt = []
        random_row_idx = self._random_row_idx_helper(data)

        while not done:
            pt = copy.copy(self)

            for col_name in col_to_var.keys():
                col_index = columns.index(col_name)
                variables = col_to_var[col_name]

                for idx, var in enumerate(variables):
                    top = len(data) - 1
                    index = pt._get_idx(strategy, start_index, idx, top, random_row_idx)
                    row = data[index]
                    value = row[col_index]
                    pt.sub(var, str(value))

                current_row_idx = start_index + idx
                if (strategy == "sequential" or strategy == "sample") and current_row_idx >= len(data):
                    done = True
                    break

            if not done:
                complete_prompt.append(pt)

            if n > -1 and len(complete_prompt) >= n:
                done = True

            start_index += len(variables)

        return self._return_single_prompt_or_completed_list(complete_prompt, n)

    def sub_from_csv(
        self,
        csv_path: str,
        col_to_var: Union[dict, Literal["infer"]] = "infer",
        headers: bool = True,
        strategy: Literal["random", "sequential", "sample"] = "sequential",
        start_index: int = 0,
        n: int = 1,
        encoding: str = "utf-8-sig",
        delimiter: str = ",",
    ) -> Union["PromptPattern", list["PromptPattern"]]:
        """
        This will replace every variable in the col_to_var dict using data from the provided file.

        Args:
            csv_path (str): Path to csv file.
            col_to_var (Union[dict, Literal["infer"]]): "infer" string or dict of column name to list of variables.
                If "infer", the col_to_var mapping is inferred based on the template's variables' name.
                For example, assuming a variable "name" and a column "name" in csv file, "infer"
                is equivalent to col_to_var={"name":["name"]}. Defaults to "infer". In its most
                general form, a user can specify a mapping like col_to_var={"column": list_of_vars,. ..}
                where the list against each column is a list of variables which can be substituted from
                that column.
            headers (bool, Optional): Indicates if the csv contains a header row. Only used if inferring col_to_var.
                Defaults to True.
            strategy (Literal[sequential, sample, random], optional): Define how the template will be populated from
                the given csv.
                Sequential  : sequentially subs the template's variables from the json, starting at a given start_index.
                Sample      : randomly samples rows and instantiate prompts. Without repetition of the data.
                Random      : for each prompt variable, substitute the value from the corresponding key at random
                Defaults to "sequential".
            start_index (int, optional): Only used with strategy sequential. State the first row to be used for the
                substitution in the template. Defaults to 0.
            n (int, Optional): Desired number of prompts to be returned.
                - n = 1: default, returns a PromptPattern
                - n > 1: returns a list of PromptPattern up to min(n, maximum producible prompts from file)
                - n = -1: goes through the entire file.
                If n is -1 and strategy is Random, a GenAiException is raised.
            encoding (str, Optional): File encoding. Defaults to "utf-8-sig".
            delimiter (str, Optional): Delimiter for the csv. Defaults to ",".

        Returns:
            PromptPattern | list[PromptPattern]: A populated PromptPattern or a list of PromptPatterns
        """

        if strategy == "random" and n <= 0:
            raise GenAiException("n must be at least 1")

        with open(csv_path, "r", encoding=encoding) as file:
            reader = csv.reader(file, delimiter=delimiter)
            list_of_rows = list(reader)
            if len(list_of_rows) < 1:
                raise ValueError("CSV file seems to be empty: {}".format(csv_path))
            if col_to_var == "infer":
                col_to_var, columns = self._tabular_infer_mode_helper(list_of_rows[0])
                relevant_rows = list_of_rows[1:] if headers else list_of_rows
                data = [row for row in relevant_rows if len(row) >= len(columns)]
            else:
                columns = list_of_rows[0]
                data = [row for row in list_of_rows[1:] if len(row) >= len(columns)]

            if strategy == "sample":
                start_index = 0

            PromptPattern.validate_start_index(strategy, start_index, data)

            complete_pt = self._sub_from_tabular_data(data, columns, col_to_var, start_index, n, strategy)

        return complete_pt

    def render(self, inputs, data):
        if not self.watsonx:
            raise GenAiException("Method only available for watsonx prompt templates.")

        _data = {"value": self.watsonx.value, "data": data}

        return PromptTemplateManager.render_watsonx_prompts(credentials=self.credentials, inputs=inputs, data=_data)

    def delete(self) -> str:
        if not self.watsonx:
            raise GenAiException("Method only available for watsonx prompt templates.")
        return PromptTemplateManager.delete_template(credentials=self.credentials, id=self.watsonx.id)

    def __str__(self):
        return self.dump

    def __repr__(self):
        return self.dump

    def __contains__(self, val):
        return self.__dict__.__contains__(val)

    @staticmethod
    def list_str(list_of_prompts: Union[List["PromptPattern"], List[str]]) -> list[str]:
        return [str(x) for x in list_of_prompts]

    @staticmethod
    def validate_start_index(strategy, start_index, data):
        if strategy == "sequential" and start_index >= len(data):
            raise GenAiException("start_index cannot be greater than the length of the file.")



from genai.schemas.descriptions import Descriptions
from genai.schemas.files_params import FileListParams, MultipartFormData
from genai.schemas.generate_params import (
    GenerateParams,
    LengthPenalty,
    Return,
    ReturnOptions,
)
from genai.schemas.history_params import HistoryParams
from genai.schemas.responses import GenerateResult, TokenizeResult
from genai.schemas.token_params import TokenParams
from genai.schemas.tunes_params import (
    CreateTuneHyperParams,
    CreateTuneParams,
    TunesListParams,
)

__all__ = [
    "Descriptions",
    "GenerateParams",
    "LengthPenalty",
    "Return",
    "ReturnOptions",
    "TokenParams",
    "HistoryParams",
    "GenerateResult",
    "TokenizeResult",
    "FileListParams",
    "MultipartFormData",
    "TunesListParams",
    "CreateTuneParams",
    "CreateTuneHyperParams",
]



import os

from dotenv import load_dotenv

from genai.credentials import Credentials
from genai.model import Model
from genai.schemas import GenerateParams

# make sure you have a .env file under genai root with
# GENAI_KEY=<your-genai-key>
load_dotenv()
api_key = os.getenv("GENAI_KEY", None)
api_endpoint = os.getenv("GENAI_API", None)
creds = Credentials(api_key, api_endpoint)  # credentials object to access GENAI

# Instantiate parameters for text generation
params = GenerateParams(decoding_method="sample", max_new_tokens=10)

# model object
flan = Model("google/flan-ul2", params=params, credentials=creds)
ul2 = Model("google/ul2", params=params, credentials=creds)
t5 = Model("google/flan-t5-xl", params=params, credentials=creds)

# Q&A with flan-ul2
print("\n------------- Example (Input flan-ul2 model as string)-------------\n")
prompts = ["What is your name?", "Where are you from?"]
for response in flan.generate_async(prompts):
    print(f"Prompt: {response.input_text}\nResponse: {response.generated_text}")

# Q&A with ul2
print("\n------------- Example (Input ul2 model as string)-------------\n")
for response in ul2.generate_async(prompts):
    print(f"Prompt: {response.input_text}\nResponse: {response.generated_text}")

# Q&A with gpt
print("\n------------- Example (Input t5-xl model as string)-------------\n")
for response in t5.generate_async(prompts, throw_on_error=True):
    print(f"Prompt: {response.input_text}\nResponse: {response.generated_text}")



import os
from pathlib import Path

from dotenv import load_dotenv

from genai.credentials import Credentials
from genai.schemas import FileListParams
from genai.services import FileManager

load_dotenv()
API_KEY = os.getenv("GENAI_KEY", None)
ENDPOINT = os.getenv("GENAI_API", None)

creds = Credentials(api_key=API_KEY, api_endpoint=ENDPOINT)
params = FileListParams(limit=5, offset=0)

# LIST FILES
print("-------- Listing files ----------")
file_list = FileManager.list_files(credentials=creds, params=params)
for f in file_list.results:
    print(f, "\n")

# GET METADATA
file_id = file_list.results[0].id
print("-------- File metadata with id = {} --------".format(file_id))
file_metadata = FileManager.file_metadata(credentials=creds, file_id=file_id)
print("File Metadada: \n", file_metadata)

# READ FILE
print("------- Read a file with id = {} -----".format(file_id))
file_content = FileManager.read_file(credentials=creds, file_id=file_id)
print("File content: \n", file_content)

# # UPLOAD FILE
# file_path = "<path-to-file>"
file_path = str(Path(__file__).parent / "file_to_tune.jsonl")
print("------- Uploading local file {} ------".format(file_path))
upload_file = FileManager.upload_file(credentials=creds, file_path=file_path, purpose="tune")
print("File uploaded: \n", upload_file)

file_list = FileManager.list_files(credentials=creds, params=params)
for f in file_list.results:
    if f.file_name == "file_to_tune.jsonl":
        file_id = f.id
        break
print("Uploaded file has id = {}\n".format(file_id))

# DELETE FILE
file_id_to_delete = "<some-file-id>"
print("-------- Deleting the above uploaded file -------")
file_delete = FileManager.delete_file(credentials=creds, file_id=file_id)
print("RESPONSE DELETE: \n", file_delete)



import os
import pathlib
import random

import requests

import genai.extensions.huggingface  # noqa
from genai.prompt_pattern import PromptPattern

#
# Adult dataset in the following example was obtained from UCI Machine
# Learning repository. The full Bibtex citation is as follows:
#
# @misc{Dua:2019 ,
# author = "Dua, Dheeru and Graff, Casey",
# year = "2017",
# title = "{UCI} Machine Learning Repository",
# url = "http://archive.ics.uci.edu/ml",
# institution = "University of California, Irvine, School of Information and Computer Sciences" }


def get_dataset(datafile):
    url = "https://archive.ics.uci.edu/ml/machine-learning-databases/adult/adult.data"

    if os.path.exists(datafile):
        print("Data already downloaded. Skipping the download.")
        return
    with requests.get(url) as response:
        print("Saving data to {}".format(datafile))
        # Add data
        with open(datafile, "ab") as f:
            f.write(response.content)


datafile = os.path.join(pathlib.Path(__file__).parent.absolute(), "adult.data.csv")
get_dataset(datafile)

# flake8: noqa
template = """
    The age of the individual is {{0}}.
    The workclass of the individual is {{1}}.
    The estimate of the number of individuals in the population with the same demographics as the individual was {{2}}.
    The numeric form of the highest education level they achieved was {{4}}.
    The marital status of the individual was {{5}}.
    The occupation of the individual was {{6}}.
    The relationship status of the individual is {{7}}.
    The race of the individual was {{8}}.
    The sex of the individual is {{9}}.
    The capital gain they had in the previous year was {{10}}.
    The capital loss they had in the previous year was {{11}}.
    The number of hours worked per week was {{12}}.
    The native country of the individual was {{13}}.
"""


prompt = PromptPattern.from_str(template)

list_of_prompts = prompt.sub_all_from_csv(csv_path=datafile, col_to_var="infer", headers=False)

print("-----------------------")
print("Generated prompts: \n total number {}".format(len(list_of_prompts)))
print("Sample prompt: {}".format(list_of_prompts[random.randint(0, len(list_of_prompts) - 1)]))
print("-----------------------")

prompt.huggingface.save_dataset(list_of_prompts, "<path-to-your-folder>")



import re
from unittest.mock import patch

import pytest
from _pytest.fixtures import SubRequest
from pytest_httpx import HTTPXMock

from genai.services import AsyncResponseGenerator, ServiceInterface


@pytest.fixture
def assert_all_responses_were_requested() -> bool:
    return False


@pytest.fixture
def non_mocked_hosts() -> list:
    return ["0.0.0.0"]


@pytest.fixture(autouse=True, scope="function")
def patch_generate_limits(request: SubRequest, httpx_mock: HTTPXMock):
    if "integration" in request.node.keywords:
        yield
    else:
        custom_params = request.param if hasattr(request, "param") else {}
        token_capacity = custom_params.get("tokens_capacity", 10)
        tokens_used = custom_params.get("tokens_used", 0)

        httpx_mock.add_response(
            url=re.compile(f".+{ServiceInterface.GENERATE_LIMITS}$"),
            method="GET",
            json={"tokenCapacity": token_capacity, "tokensUsed": tokens_used},
            status_code=200,
        )
        yield


@pytest.fixture(autouse=True, scope="function")
def patch_async_requests_limits(request: SubRequest):
    if "integration" in request.node.keywords:
        yield
        return

    custom_params = request.param if hasattr(request, "param") else {}
    sleep_duration = custom_params.get("duration", 0.1)
    with patch.multiple(AsyncResponseGenerator, LIMITS_CHECK_SLEEP_DURATION=sleep_duration):
        yield



import logging
import os
from collections import OrderedDict

from genai.credentials import Credentials
from genai.exceptions.genai_exception import GenAiException
from genai.schemas import FileListParams
from genai.schemas.responses import FileInfoResult, FilesListResponse
from genai.services import ServiceInterface
from genai.utils.errors import to_genai_error
from genai.utils.service_utils import _get_service

logger = logging.getLogger(__name__)


class FileManager:
    """Class for managing files on the server."""

    @staticmethod
    def list_files(
        credentials: Credentials = None,
        service: ServiceInterface = None,
        params: FileListParams = None,
    ) -> FilesListResponse:
        """List all files on the server.

        Args:
            credentials (Credentials, optional): Credentials object. Defaults to None.
                If not providec, service must be provided.
            service (ServiceInterface, optional): ServiceInterface object. Defaults to None.
                If not provided, credentials must be provided.
            params (FileListParams, optional): Parameters for listing files.

        Returns:
            FilesListResponse: Response from the server.
        """
        service = _get_service(credentials, service)

        try:
            response = service._files.list_files(params=params)

            if response.status_code == 200:
                response = response.json()
                responses = FilesListResponse(**response)
                return responses
            else:
                raise GenAiException(response)
        except Exception as e:
            raise to_genai_error(e)

    @staticmethod
    def file_metadata(
        file_id: str, credentials: Credentials = None, service: ServiceInterface = None
    ) -> FileInfoResult:
        """Get metadata from a file.

        Args:
            credentials (Credentials, optional): Credentials object. Defaults to None.
                If not providec, service must be provided.
            service (ServiceInterface, optional): ServiceInterface object. Defaults to None.
                If not provided, credentials must be provided.
            file_id (str): File id.

        Returns:
            FileInfoResult: Response from the server.
        """
        service = _get_service(credentials, service)

        try:
            response = service._files.get_file_metadata(file_id=file_id)

            if response.status_code == 200:
                response = response.json()
                return FileInfoResult(**response["results"])
            else:
                raise GenAiException(response)
        except Exception as e:
            raise to_genai_error(e)

    @staticmethod
    def read_file(file_id: str, credentials: Credentials = None, service: ServiceInterface = None) -> str:
        """Read a file from the server and return the file content.

        Args:
            file_id (str): File id or list of file ids.
            credentials (Credentials, optional): Credentials object. Defaults to None.
                If not provided, service must be provided.
            service (ServiceInterface, optional): ServiceInterface object. Defaults to None.
                If not provided, credentials must be provided.

        Returns:
            str: File content.
        """
        service = _get_service(credentials, service)

        try:
            response = service._files.read_file(file_id=file_id)

            if response.status_code == 200:
                return response.content.decode("utf-8")
            else:
                raise GenAiException(response)
        except Exception as e:
            raise to_genai_error(e)

    @staticmethod
    def upload_file(
        file_path: str,
        purpose: str,
        credentials: Credentials = None,
        service: ServiceInterface = None,
    ) -> FileInfoResult:
        """Upload a file to the server.

        Args:
            file_path (str): Path to the file to be uploaded. The file needs to be in JSON or JSON Lines format.
                Format of the data depends on the task_id.
            purpose (str): Purpose of the file to be uploaded. Currently accepts only "tune" or "template".
            credentials (Credentials, optional): Credentials object. Defaults to None.
                If not providec, service must be provided.
            service (ServiceInterface, optional): ServiceInterface object. Defaults to None.
                If not provided, credentials must be provided.

        Raises:
            GenAiException: If file does not exist or has incorrect format or if incorrect purpose.

        Returns:
            FileInfoResult: Response from the server.
        """

        if purpose != "tune" and purpose != "template":
            raise GenAiException(ValueError("Purpose must be 'tune' or 'template'"))

        # check if file exists
        if not os.path.isfile(file_path):
            raise GenAiException(ValueError(f"File {file_path} does not exist"))

        # check if file is json or jsonl
        if not file_path.endswith(".json") and not file_path.endswith(".jsonl"):
            raise GenAiException(ValueError(f"File {file_path} must be in json or jsonl format"))

        multipart_form_data = {
            "purpose": (None, purpose),
            "file": (file_path, open(file_path, "rb")),
        }

        multipart_form_data = FileManager._validate_mmultipart_form_data_order(multipart_form_data)

        service = _get_service(credentials, service)

        try:
            response = service._files.upload_file(multipart_form_data=multipart_form_data)

            if response.status_code == 201:
                response = response.json()
                return FileInfoResult(**response["results"])
            else:
                raise GenAiException(response)
        except Exception as e:
            raise to_genai_error(e)

    @staticmethod
    def delete_file(file_id: str, credentials: Credentials = None, service: ServiceInterface = None) -> dict:
        """Delete a file from the server.

        Args:
            file_id (str): File id.
            credentials (Credentials, optional): Credentials object. Defaults to None.
                If not providec, service must be provided.
            service (ServiceInterface, optional): ServiceInterface object. Defaults to None.
                If not provided, credentials must be provided.

        Returns:
            dict: Response from the server.
        """
        service = _get_service(credentials, service)

        try:
            response = service._files.delete_file(file_id=file_id)
            if response.status_code == 204:
                return {"status": "success"}
            else:
                raise GenAiException(response)
        except Exception as e:
            raise to_genai_error(e)

    @staticmethod
    def _validate_mmultipart_form_data_order(form_data: dict) -> dict:
        """
        To upload a file, the required body needs to be in a specific order.
        'purpose' needs to be placed before 'file' in the multipart/form-data payload.
        Args:
            form_data (dict): Dictionary with the form data.
        Returns:
            dict: Dictionary with the form data in the correct order.
        """
        desired_order = ["purpose", "file"]

        if list(form_data.keys()) != desired_order:
            # Reorder the dictionary using an OrderedDict
            form_data = OrderedDict((key, form_data.get(key)) for key in desired_order if key in form_data)
        return form_data



import pytest
from pydantic import ValidationError

from genai.schemas import GenerateParams, LengthPenalty, Return, ReturnOptions

# API Reference : https://workbench.res.ibm.com/docs/api-reference#generate


@pytest.mark.unit
class TestGenerateSchema:
    def setup_method(self):
        # mock object for the API call
        self.model = "google/ul2"
        self.inputs = ["Write a tagline for an alumni association: Together we"]

        # test all GenerateParams fields
        self.params = GenerateParams(
            decoding_method="sample",
            length_penalty=LengthPenalty(decay_factor=1.5, start_index=2),
            max_new_tokens=3,
            min_new_tokens=1,
            random_seed=123,
            stop_sequences=["!", "?", ".", "\n"],
            stream=False,
            temperature=0.5,
            time_limit=10.0,
            top_k=10,
            top_p=0.7,
            typical_p=0.5,
            repetition_penalty=1.2,
            truncate_input_tokens=2,
            beam_width=1,
            return_options=ReturnOptions(
                input_text=True,
                generated_tokens=True,
                input_tokens=True,
                token_logprobs=False,
            ),
        )

    def test_preserve_special_characters(self):
        assert "\n" in self.params.stop_sequences

    @pytest.fixture
    def request_body(self):
        return {
            "model": self.model,
            "inputs": self.inputs,
            "params": self.params,
        }

    def test_existing_required_fields(self, request_body):
        assert "model" in request_body
        assert "inputs" in request_body

    def test_model(self, request_body):
        # test that the models list is not empty
        assert len(self.model) > 0
        # test that the models list contains only strings
        assert all(isinstance(model, str) for model in self.model)

    def test_inputs(self, request_body):
        # test that the inputs list is not empty
        assert len(self.inputs) > 0
        # test that the inputs list contains only strings
        assert all(isinstance(input, str) for input in self.inputs)

    def test_invalid_fields(self):
        # test generate invalid fields
        with pytest.raises(ValidationError):
            GenerateParams(
                decoding_method="invalid_method",
                length_penalty=LengthPenalty(decay_factor=0.5, start_index=-1),
                max_new_tokens=0,
                min_new_tokens=20,
                random_seed=10000,
                stop_sequences=[""],
                stream="false",
                temperature=-0.5,
                time_limit=-1.0,
                top_k=0,
                top_p=1.5,
                repetition_penalty=-0.5,
                truncate_input_tokens=-2.0,
                return_options=ReturnOptions(input_text="true", generated_tokens="false"),
            )

    def test_decoding_method_invalid_type(self):
        with pytest.raises(ValidationError):
            GenerateParams(decoding_method="invalid_method")

    def test_decoding_method_valid_type(self, request_body):
        # test that decoding_method must be either greedy or sample
        params = request_body["params"]
        assert params.decoding_method in ["greedy", "sample"]

    def test_max_new_tokens_invalid_type(self):
        # test that max_new_tokens must be an integer
        with pytest.raises(ValidationError):
            GenerateParams(max_new_tokens="3.5")
        with pytest.raises(ValidationError):
            GenerateParams(max_new_tokens="dummy")

    def test_max_new_token_valid_type(self, request_body):
        # test that max_new_tokens must be an intenger between 1 and 1024
        params = request_body["params"]
        assert params.max_new_tokens >= 1
        assert params.max_new_tokens <= 1024
        assert isinstance(params.max_new_tokens, int)

    def test_min_new_tokens_invalid_type(self):
        # test that min_new_tokens must be an integer
        with pytest.raises(ValidationError):
            GenerateParams(min_new_tokens="2.5")

    def test_min_new_tokens_valid_type(self, request_body):
        # test that min_new_tokens must be an integer and greater than 0
        # NOTE: min_new_tokens must be less than max_new_tokens?
        # NOTE: min_new_tokens can be 0?
        params = request_body["params"]
        assert params.min_new_tokens > 0
        assert isinstance(params.min_new_tokens, int)

    def test_random_seed_invalid_type(self):
        # test that random_seed must be an integer between 1 and 9999
        with pytest.raises(ValidationError):
            GenerateParams(random_seed="dummy")
        with pytest.raises(ValidationError):
            GenerateParams(random_seed=0)

    def test_random_seed_valid_type(self, request_body):
        # test that random_seed must be an integer between 1 and 9999
        params = request_body["params"]
        assert params.random_seed >= 1
        assert params.random_seed <= 9999
        assert isinstance(params.random_seed, int)

    def test_stop_sequences_invalid_type(self):
        # test that stop_sequences must be a list of strings and not empty
        with pytest.raises(ValidationError):
            GenerateParams(stop_sequences=".")
        with pytest.raises(ValidationError):
            GenerateParams(stop_sequences=42)

    def test_stop_sequences_valid_type(self, request_body):
        # test that stop_sequences must be a list of strings and not empty
        params = request_body["params"]
        assert len(params.stop_sequences) > 0
        assert all(isinstance(stop, str) for stop in params.stop_sequences)

    def test_stream_invalid_type(self):
        # test that stream must be a boolean
        with pytest.raises(ValidationError):
            GenerateParams(stream="")
        with pytest.raises(ValidationError):
            GenerateParams(stream=[0, 1, 2])
        with pytest.raises(ValidationError):
            GenerateParams(stream="dummy")
        with pytest.raises(ValidationError):
            GenerateParams(stream=123)

    def test_stream_valid_type(self, request_body):
        # test that stream must be a boolean
        params = request_body["params"]
        assert isinstance(params.stream, bool)

    def test_temperature_invalid_type(self):
        # test that temperature must be a float between 0.05 and 1
        with pytest.raises(ValidationError):
            GenerateParams(temperature="")
        with pytest.raises(ValidationError):
            GenerateParams(temperature=[0, 1, 2])
        with pytest.raises(ValidationError):
            GenerateParams(temperature="dummy")
        with pytest.raises(ValidationError):
            GenerateParams(temperature=-0.5)
        with pytest.raises(ValidationError):
            GenerateParams(temperature=2.5)

    def test_temperature_valid_type(self, request_body):
        # test that temperature must be a float between 0 and 1
        params = request_body["params"]
        assert params.temperature >= 0
        assert params.temperature <= 2
        assert isinstance(params.temperature, float)

    def test_time_limit_invalid_type(self):
        # test that time_limit must be a int and greater than 0
        # NOTE: time_limit can be 0 or less then 0?
        with pytest.raises(ValidationError):
            GenerateParams(time_limit="")
        with pytest.raises(ValidationError):
            GenerateParams(time_limit=[0, 1, 2])
        with pytest.raises(ValidationError):
            GenerateParams(time_limit="dummy")

    def test_time_limit_valid_type(self, request_body):
        # test that time_limit must be a int
        params = request_body["params"]
        assert isinstance(params.time_limit, int)
        assert params.time_limit >= 0

    def test_top_k_invalid_type(self):
        # test that top_k must be an integer and greater than 0
        with pytest.raises(ValidationError):
            GenerateParams(top_k="")
        with pytest.raises(ValidationError):
            GenerateParams(top_k=[0, 1, 2])
        with pytest.raises(ValidationError):
            GenerateParams(top_k="dummy")
        with pytest.raises(ValidationError):
            GenerateParams(top_k=0.5)
        with pytest.raises(ValidationError):
            GenerateParams(top_k=0)

    def test_top_k_valid_type(self, request_body):
        # test that top_k must be an integer and greater than 0
        params = request_body["params"]
        assert isinstance(params.top_k, int)
        assert params.top_k > 0

    def test_top_p_invalid_type(self):
        # test that top_p must be a float between 0 and 1
        with pytest.raises(ValidationError):
            GenerateParams(top_p="")
        with pytest.raises(ValidationError):
            GenerateParams(top_p=[0, 1, 2])
        with pytest.raises(ValidationError):
            GenerateParams(top_p="dummy")
        with pytest.raises(ValidationError):
            GenerateParams(top_p=-0.5)
        with pytest.raises(ValidationError):
            GenerateParams(top_p=1.5)

    def test_top_p_valid_type(self, request_body):
        # test that top_p must be a float between 0 and 1
        params = request_body["params"]
        assert params.top_p >= 0
        assert params.top_p <= 1
        assert isinstance(params.top_p, float)

    def test_typical_p_invalid_type(self):
        with pytest.raises(ValidationError):
            GenerateParams(typical_p="")
        with pytest.raises(ValidationError):
            GenerateParams(typical_p=[0, 1, 2])
        with pytest.raises(ValidationError):
            GenerateParams(typical_p="dummy")
        with pytest.raises(ValidationError):
            GenerateParams(typical_p=0)
        with pytest.raises(ValidationError):
            GenerateParams(typical_p=1.5)

    def test_typical_p_valid_type(self, request_body):
        params = request_body["params"]
        assert params.typical_p > 0
        assert params.typical_p <= 1
        assert isinstance(params.typical_p, float)

    def test_repetition_penalty_invalid_type(self):
        # test that repetition_penalty must be a float
        # NOTE: repetition_penalty can be 0 or less than 0?
        with pytest.raises(ValidationError):
            GenerateParams(repetition_penalty="")
        with pytest.raises(ValidationError):
            GenerateParams(repetition_penalty=[0, 1, 2])
        with pytest.raises(ValidationError):
            GenerateParams(repetition_penalty="dummy")

    def test_repetition_penalty_valid_type(self, request_body):
        # test that repetition_penalty must be a float
        params = request_body["params"]
        # assert params.repetition_penalty >= 0
        assert isinstance(params.repetition_penalty, float)

    def test_truncate_input_tokens_invalid_type(self):
        # test that truncate_input_tokens must be an integer
        with pytest.raises(ValidationError):
            GenerateParams(truncate_input_tokens="")
        with pytest.raises(ValidationError):
            GenerateParams(truncate_input_tokens=[0, 1, 2])
        with pytest.raises(ValidationError):
            GenerateParams(truncate_input_tokens="dummy")

    def test_beam_width_valid_type(self, request_body):
        params = request_body["params"]
        assert isinstance(params.beam_width, int)

    def test_beam_width_invalid_type(self):
        # test that beam_width must be an non-negative integer
        with pytest.raises(ValidationError):
            GenerateParams(beam_width="")
        with pytest.raises(ValidationError):
            GenerateParams(beam_width=-100)

    def test_truncate_input_tokens_valid_type(self, request_body):
        params = request_body["params"]
        assert isinstance(params.truncate_input_tokens, int)

    def test_returns_raises_warning(self):
        with pytest.deprecated_call():
            GenerateParams(returns=Return())

    def test_optional_fields(self):
        try:
            genparams = GenerateParams(answer=42)
            assert genparams.answer == 42
        except ValidationError as e:
            assert False, "Extra fields not parsed: {}".format(e)



from typing import Optional

from pydantic import BaseModel, Extra, Field

from genai.schemas import Descriptions as tx

# API Reference : https://workbench.res.ibm.com/docs


class TokenParams(BaseModel):
    class Config:
        anystr_strip_whitespace = True
        extra = Extra.forbid

    return_tokens: Optional[bool] = Field(None, description=tx.RETURN_TOKEN)



import logging

from genai.credentials import Credentials
from genai.exceptions import GenAiException
from genai.schemas.history_params import HistoryParams
from genai.schemas.responses import HistoryResponse, TermsOfUse
from genai.services import ServiceInterface
from genai.utils.errors import to_genai_error

logger = logging.getLogger(__name__)


class Metadata:
    DEFAULT_MAX_PROMPTS = 5

    def __init__(self, credentials: Credentials):
        """Gets metadata for service backing the Model class.

        Args:
            credentials (Credentials): The API Credentials
        """

        logger.debug(f"Metadata Created: api_endpoint: {credentials.api_endpoint}")
        self.service = ServiceInterface(service_url=credentials.api_endpoint, api_key=credentials.api_key)

    def accept_terms_of_use(self) -> TermsOfUse:
        """Accepts the terms of use on GENAI.

        Returns:
            TermsOfUse: Terms of Use Data
        """

        try:
            tou_response = self.service.terms_of_use(True)

            if tou_response.status_code == 200:
                tou_data = TermsOfUse(**tou_response.json())
                return tou_data
            else:
                raise GenAiException(tou_response)
        except Exception as ex:
            raise to_genai_error(ex)

    def get_history(self, params: HistoryParams = HistoryParams()) -> HistoryResponse:
        """The requests endpoint provides the ability to retrieve past generation requests
           and responses returned by the given models.
           Items are returned in reverse chronological order.

        Args:
            params (HistoryParams): History parameters

        Returns:
            HistoryResponse: The History of requests and responses
        """

        try:
            history_response = self.service.history(params)

            if history_response.status_code == 200:
                history_data = HistoryResponse(**history_response.json())
                return history_data
            else:
                raise GenAiException(history_response)
        except Exception as ex:
            raise to_genai_error(ex)



import logging
import os
import pathlib
from typing import Optional

from genai.credentials import Credentials
from genai.exceptions.genai_exception import GenAiException
from genai.schemas.responses import (
    TuneGetResponse,
    TuneInfoResult,
    TuneMethodsGetResponse,
    TunesListResponse,
)
from genai.schemas.tunes_params import (
    CreateTuneParams,
    DownloadAssetsParams,
    TunesListParams,
)
from genai.services.service_interface import ServiceInterface
from genai.utils.errors import to_genai_error
from genai.utils.service_utils import _get_service

logger = logging.getLogger(__name__)


class TuneManager:
    """Class for managing tunes on the server."""

    @staticmethod
    def list_tunes(
        credentials: Credentials = None,
        service: ServiceInterface = None,
        params: TunesListParams = None,
    ) -> TunesListResponse:
        """List all tunes on the server.

        Args:
            credentials (Credentials, optional): Credentials object. Defaults to None.
                If not providec, service must be provided.
            service (ServiceInterface, optional): ServiceInterface object. Defaults to None.
                If not provided, credentials must be provided.
            params (TunesListParams): Parameters for listing tunes.

        Returns:
            TunesListResponse: Response from the server.
        """
        service = _get_service(credentials, service)
        try:
            response = service._tunes.list_tunes(params=params)

            if response.status_code == 200:
                response = response.json()
                responses = TunesListResponse(**response)
                return responses
            else:
                raise GenAiException(response)
        except Exception as e:
            raise to_genai_error(e)

    @staticmethod
    def get_tune(
        tune_id: str, credentials: Credentials = None, service: ServiceInterface = None
    ) -> Optional[TuneInfoResult]:
        """Get a tune from the server.

        Args:
            tune_id: ID of the tune to fetch from the server.
            credentials (Credentials, optional): Credentials object. Defaults to None.
                If not providec, service must be provided.
            service (ServiceInterface, optional): ServiceInterface object. Defaults to None.
                If not provided, credentials must be provided.

        Returns:
            TuneInfoResult: Response from the server.
        """
        service = _get_service(credentials, service)
        try:
            response = service._tunes.get_tune(tune_id=tune_id)
            if response.status_code == 200:
                response = response.json()
                responses = TuneGetResponse(**response)
                return responses.results
            else:
                raise GenAiException(response)
        except Exception as e:
            raise to_genai_error(e)

    @staticmethod
    def create_tune(
        params: CreateTuneParams,
        credentials: Credentials = None,
        service: ServiceInterface = None,
    ) -> TuneInfoResult:
        """Create a new tune to be uploaded to the server.

        Args:
            params (CreateTuneParams): Parameters for creating tunes.
            credentials (Credentials, optional): Credentials object. Defaults to None.
                If not providec, service must be provided.
            service (ServiceInterface, optional): ServiceInterface object. Defaults to None.
                If not provided, credentials must be provided.

        Returns:
            TuneInfoResult: Response from the server.
        """

        if params.method_id == "mpt" and (params.parameters.init_method or params.parameters.init_text):
            raise GenAiException(
                "When using method_id 'mpt' you cannot use init_method or "
                "init_text, those are only for 'pt' (Prompt Tuning) method_id."
            )

        if params.task_id not in ["generation", "classification", "summarization"]:
            raise GenAiException(
                "The task_id must be one of the following: 'generation', 'classification' or 'summarization'."
            )

        service = _get_service(credentials, service)
        try:
            response = service._tunes.create_tune(params=params)
            if response.status_code == 200:
                response = response.json()
                responses = TuneInfoResult(**response["results"])
                return responses
            else:
                raise GenAiException(response)
        except Exception as e:
            raise to_genai_error(e)

    @staticmethod
    def delete_tune(tune_id: str, credentials: Credentials = None, service: ServiceInterface = None) -> dict:
        """Deletes a tune from the server.

        Args:
            tune_id: ID of the tune to delete from the server.
            credentials (Credentials, optional): Credentials object. Defaults to None.
                If not providec, service must be provided.
            service (ServiceInterface, optional): ServiceInterface object. Defaults to None.
                If not provided, credentials must be provided.

        Returns:
            dict: Response from the server.
        """
        service = _get_service(credentials, service)
        status = TuneManager._get_tune_status(tune_id, service)

        if status not in ["FAILED", "COMPLETED"]:
            raise Exception(f"Tune status: {status}. Tune can not be deleted until is completed or failed.")

        try:
            response = service._tunes.delete_tune(tune_id=tune_id)
            if response.status_code == 204:
                return {"status": "success"}
            else:
                raise GenAiException(response)
        except Exception as e:
            raise to_genai_error(e)

    @staticmethod
    def get_tune_methods(credentials: Credentials = None, service: ServiceInterface = None) -> TuneMethodsGetResponse:
        """Get list of tune methods.

        Args:
            credentials (Credentials, optional): Credentials object. Defaults to None.
                If not providec, service must be provided.
            service (ServiceInterface, optional): ServiceInterface object. Defaults to None.
                If not provided, credentials must be provided.

        Returns:
            TuneMethodsGetResponse: Response from the server
        """
        service = _get_service(credentials, service)
        try:
            response = service._tunes.get_tune_methods()
            if response.status_code == 200:
                response = response.json()
                responses = TuneMethodsGetResponse(**response)
                return responses
            else:
                raise GenAiException(response)
        except Exception as e:
            raise to_genai_error(e)

    @staticmethod
    def get_filename(params: DownloadAssetsParams):
        if params.content == "encoder":
            return f"{params.id}.pt"
        elif params.content == "logs":
            return f"{params.id}.jsonl"
        else:
            raise GenAiException("Input for 'content' must either be 'encoder' or 'logs'")

    @staticmethod
    def get_complete_path(output_path, filename):
        if output_path == "tune_assets":
            parent_path = pathlib.Path(__file__).parent.resolve()
            output_path = os.path.join(parent_path, output_path)
        if not (os.path.exists(output_path) and os.path.isdir(output_path)):
            os.makedirs(output_path)
        return os.path.join(output_path, filename)

    @staticmethod
    def _get_tune_status(tune_id, service):
        """Get status of a tuned model."""
        try:
            response = service._tunes.get_tune(tune_id=tune_id)
            if response.status_code == 200:
                response = response.json()
                return response["results"]["status"]
        except Exception as e:
            raise to_genai_error(e)

    @staticmethod
    def download_tune_assets(
        params: DownloadAssetsParams,
        output_path="tune_assets",
        credentials: Credentials = None,
        service: ServiceInterface = None,
    ) -> dict:
        """Download tune asset (available only for completed tunes)

        Args:
            params (DownloadAssetsParams): Parameters for downloading tune assets.
            output_path (str): Output directory for tune assets file. Path is relative to services directory.
            credentials (Credentials, optional): Credentials object. Defaults to None.
                If not providec, service must be provided.
            service (ServiceInterface, optional): ServiceInterface object. Defaults to None.
                If not provided, credentials must be provided.
        Returns:
            dict: Response from the server.
        """

        filename = TuneManager.get_filename(params)
        service = _get_service(credentials, service)
        status = TuneManager._get_tune_status(params.id, service)

        # while status not in ["FAILED", "HALTED", "COMPLETED"]:
        #     logger.debug(f"Tune status: {status}. Tune can not be dowloaded until is completed.")
        if status != "COMPLETED":
            raise GenAiException(f"Tune status: {status}. Tune can not be dowloaded if status is not COMPLETED.")

        try:
            response = service._tunes.download_tune_assets(params=params)
            if response.status_code == 200:
                path = TuneManager.get_complete_path(output_path, filename)
                with open(path, "wb") as f:
                    f.write(response.content)
                    return {"dowloaded_file_path": path}
            else:
                raise GenAiException(response)
        except Exception as e:
            raise to_genai_error(e)



import os

from dotenv import load_dotenv

from genai.credentials import Credentials
from genai.model import Model
from genai.schemas import TokenParams

# make sure you have a .env file under genai root with
# GENAI_KEY=<your-genai-key>
# GENAI_API=<genai-api-endpoint>
load_dotenv()
api_key = os.getenv("GENAI_KEY", None)
api_endpoint = os.getenv("GENAI_API", None)

print("\n------------- Example (Tokenize)-------------\n")

creds = Credentials(api_key, api_endpoint)
model = Model("google/flan-ul2", params=TokenParams, credentials=creds)

sentence = "Hello! How are you?"

tokenized_response = model.tokenize([sentence], return_tokens=True)
print(f"Tokenized response: {tokenized_response}")



import logging

# Import the ibm-generative-ai library and local server extension
from genai.extensions.localserver import CustomModel, LocalLLMServer
from genai.model import Model
from genai.schemas import GenerateParams, GenerateResult, TokenizeResult, TokenParams

# This example uses the transformers library, please install using:
# pip install transformers torch sentencepiece
try:
    from transformers import T5ForConditionalGeneration, T5Tokenizer
except ImportError:
    raise ImportError(
        """
Could not import transformers which is needed for this example.
Please install using: pip install transformers torch sentencepiece
"""
    )


logger = logging.getLogger(__name__)

# Create your custom model


class FlanT5Model(CustomModel):
    model_id = "google/flan-t5-base"

    def __init__(self):
        logger.info("Initialising my custom flan-t5-base model")
        self.tokenizer = T5Tokenizer.from_pretrained("google/flan-t5-base")
        self.model = T5ForConditionalGeneration.from_pretrained("google/flan-t5-base")
        logger.info("flan-t5-base is ready!")

    def generate(self, input_text: str, params: GenerateParams) -> GenerateResult:
        logger.info(f"Calling generate on: {input_text}")
        input_ids = self.tokenizer(input_text, return_tensors="pt").input_ids
        response = self.model.generate(input_ids, max_new_tokens=params.max_new_tokens)

        genai_response = GenerateResult(
            generated_text=self.tokenizer.decode(response[0]),
            generated_token_count=response.shape[1],
            input_token_count=input_ids.shape[1],
            stop_reason="",
        )
        logger.info(f"Response to {input_text} was: {genai_response}")

        return genai_response

    def tokenize(self, input_text: str, params: TokenParams) -> TokenizeResult:
        logger.info(f"Calling tokenize on: {input_text}")
        tokenised = self.tokenizer(input_text).input_ids
        tokens = self.tokenizer.convert_ids_to_tokens(tokenised)
        result = TokenizeResult(token_count=len(tokens))
        if params.return_tokens is True:
            result.tokens = tokens
        return result


# Instansiate the Local Server with your model
server = LocalLLMServer(models=[FlanT5Model])

# Start the server and execute your code
with server.run_locally():
    print(" > Server is started")
    # Get the credentials / connection details for the local server
    creds = server.get_credentials()

    # Instantiate parameters for text generation
    params = GenerateParams(decoding_method="sample", max_new_tokens=10)

    # Instantiate a model proxy object to send your requests
    flan_ul2 = Model(FlanT5Model.model_id, params=params, credentials=creds)

    prompts = ["Hello! How are you?", "How's the weather?"]
    for response in flan_ul2.generate_as_completed(prompts):
        print(f"Prompt: {response.input_text}\nResponse: {response.generated_text}")


print(" > Server stopped, goodbye!")



import os
import pathlib
from unittest.mock import MagicMock, patch

import pytest
from pydantic import ValidationError

from genai import Credentials
from genai.exceptions import GenAiException
from genai.routers import FilesRouter
from genai.schemas import FileListParams
from genai.schemas.responses import FileInfoResult, FilesListResponse
from genai.services.file_manager import FileManager
from tests.assets.response_helper import SimpleResponse


@pytest.mark.unit
class TestFiles:
    def setup_method(self):
        # mock object for the API call
        self.service_router = FilesRouter(service_url="SERVICE_URL", api_key="API_KEY")
        self.path = pathlib.Path(__file__).parent.resolve()
        self.asset_path = str(self.path) + os.sep + "assets" + os.sep

    @pytest.fixture
    def params(self):
        return FileListParams(limit=10, offset=0)

    @pytest.fixture
    def credentials(self):
        return Credentials("GENAI_APY_KEY")

    @pytest.fixture
    def file_to_upload(self):
        return self.asset_path + "file_to_tune.json"

    @pytest.fixture
    def not_valid_file_to_upload(self):
        return self.asset_path + "file_to_tune.csv"

    @pytest.fixture
    def multipart_form_data(self, file_to_upload):
        return {
            "purpose": (None, "test"),
            "task_id": (None, "test"),
            "file": (file_to_upload, open(file_to_upload, "rb")),
        }

    @pytest.fixture
    def task_id_types(self):
        return ["generation", "classification", "summarization"]

    # test list files function
    @patch("genai.services.RequestHandler.get")
    def test_list_files(self, mock_requests, params, credentials):
        list_response = SimpleResponse.files(params=params)
        expected_response = FilesListResponse(**list_response)

        mock_response = MagicMock(status_code=200)
        mock_response.json.return_value = list_response
        mock_requests.return_value = mock_response

        f = FileManager.list_files(credentials=credentials)
        assert f == expected_response

    @patch("genai.services.RequestHandler.get")
    def test_file_list_api_call(self, mock_requests, params):
        mock_response = MagicMock(status_code=200)
        mock_response.json.return_value = SimpleResponse.files()
        mock_requests.return_value = mock_response

        f = self.service_router.list_files(params=params)
        assert f == mock_response

    def test_file_list_api_call_with_wrong_params(self):
        with pytest.raises(GenAiException) as e:
            self.service_router.list_files(params="params")
            assert e.message == "params must be of type FileListParams."

    def test_file_list_api_call_with_wrong_params_type(self):
        with pytest.raises(GenAiException) as e:
            self.service_router.list_files(params=123)
            assert e.message == "params must be of type FileListParams."

    # test file metadata function
    @patch("genai.services.RequestHandler.get")
    def test_file_metadata_api_call(self, mock_requests):
        mock_response = MagicMock(status_code=200)
        mock_response.json.return_value = SimpleResponse.files(file_id="file_id")
        mock_requests.return_value = mock_response

        f = self.service_router.get_file_metadata(file_id="file_id")
        assert f == mock_response

    def test_file_metadata_api_call_wrong_id(self):
        with pytest.raises(GenAiException) as e:
            self.service_router.get_file_metadata(file_id=123)
            assert e.message == "File not found, file_id must be of type str"

    @patch("genai.services.RequestHandler.get")
    def test_file_metadata(self, mock_requests, credentials):
        file_metadata_response = SimpleResponse.files(file_id="file_id")
        expected_response = FileInfoResult(**file_metadata_response["results"])

        mock_response = MagicMock(status_code=200)
        mock_response.json.return_value = file_metadata_response
        mock_requests.return_value = mock_response

        file_response = FileManager.file_metadata(credentials=credentials, file_id="file_id")

        assert file_response == expected_response

    # test file read function
    @patch("genai.services.RequestHandler.get")
    def test_read_file_api_call(self, mock_requests):
        mock_response = MagicMock(status_code=200)
        mock_response.json.return_value = SimpleResponse.files(file_id="file_id")
        mock_requests.return_value = mock_response

        f = self.service_router.read_file(file_id="file_id")
        assert f == mock_response

    def test_read_file_api_call_wrong_id(self):
        with pytest.raises(Exception) as e:
            self.service_router.read_file(file_id=123)
            assert e.message == "File not found, file_id must be of type str"

    @patch("genai.services.RequestHandler.get")
    def test_read_file(self, mocker, credentials):
        response = MagicMock(status_code=200)
        mocker.return_value = response

        file_response = FileManager.read_file(credentials=credentials, file_id="file_id")

        assert file_response == response.content.decode("utf-8")

    # test file delete function
    @patch("genai.services.RequestHandler.delete")
    def test_delete_file_api_call(self, mock_requests):
        mock_response = MagicMock(status_code=204)
        mock_response.json.return_value = {"status": "success"}
        mock_requests.return_value = mock_response

        f = self.service_router.delete_file(file_id="file_id")
        assert f == mock_response

    def test_file_delete_api_call_wrong_id(self):
        with pytest.raises(Exception) as e:
            self.service_router.delete_file(file_id=123)
            assert e.message == "File not found, file_id must be of type str"

    @patch("genai.services.RequestHandler.delete")
    def test_delete_file(self, mocker, credentials):
        expected_response = {"status": "success"}
        response = MagicMock(status_code=204)
        response.json.return_value = expected_response
        mocker.return_value = response

        file_response = FileManager.delete_file(credentials=credentials, file_id="file_id")

        assert file_response == expected_response

    # test file upload function
    @patch("genai.services.RequestHandler.post")
    def test_upload_file_api_call(self, mock_requests, multipart_form_data):
        mock_response = MagicMock(status_code=200)
        mock_response.json.return_value = SimpleResponse.files(multipart_form_data=multipart_form_data)
        mock_requests.return_value = mock_response

        f = self.service_router.upload_file(multipart_form_data=multipart_form_data)
        assert f == mock_response

    def test_file_uplod_api_call_wrong_forms(self):
        with pytest.raises(GenAiException) as e:
            self.service_router.upload_file(multipart_form_data="form_data")
            assert e.message == "multipart_form_data must be of type dict"
        with pytest.raises(GenAiException) as e:
            self.service_router.upload_file(multipart_form_data="")
            assert e.message == "multipart_form_data must be of type dict"

    @patch("genai.services.RequestHandler.post")
    def test_file_upload(self, mocker, multipart_form_data, credentials, file_to_upload):
        upload_response = SimpleResponse.files(multipart_form_data=multipart_form_data)
        expected_upload_response = FileInfoResult(**upload_response["results"])

        response = MagicMock(status_code=201)
        response.json.return_value = upload_response
        mocker.return_value = response

        file_response = FileManager.upload_file(credentials=credentials, file_path=file_to_upload, purpose="tune")

        assert file_response == expected_upload_response

    def test_file_upload_with_wrong_params_type(self, credentials, file_to_upload):
        with pytest.raises(GenAiException) as e:
            FileManager.upload_file(credentials=credentials, file_path=file_to_upload, purpose=123)
            assert e.message == "purpose is not valid, must be one of tune or template"

    def test_file_upload_with_wrong_file_path(self, credentials):
        with pytest.raises(GenAiException) as e:
            FileManager.upload_file(credentials=credentials, file_path="file_path", purpose="tune")
            assert e.message == "File not found, file_path does not exist"

    def test_file_upload_is_json(self, credentials, not_valid_file_to_upload):
        with pytest.raises(GenAiException) as e:
            FileManager.upload_file(credentials=credentials, file_path=not_valid_file_to_upload, purpose="tune")
            assert e.message == "File must be in json or jsonl format"

    def test_file_list_params_schema(self):
        with pytest.raises(ValidationError) as e:
            FileListParams(limit="200", offset=0)
            assert e.message == "Ensure the limit value is less than or equal to 100"
        with pytest.raises(ValidationError) as e:
            FileListParams(limit=[100], offset="0")
            assert e.message == "Ensure limite value is a valid integer and not list"
        with pytest.raises(ValidationError) as e:
            FileListParams(limit=100, offset=[0])
            assert e.message == "Ensure offset value is a valid integer and not list"
        with pytest.raises(ValidationError) as e:
            FileListParams(limit=100, offset=0, search=["test"])
            assert e.message == "Ensure search value is a valid string"



import os

from dotenv import load_dotenv

from genai.model import Credentials, Model
from genai.schemas import GenerateParams

# make sure you have a .env file under genai root with
# GENAI_KEY=<your-genai-key>
load_dotenv()
api_key = os.getenv("GENAI_KEY", None)
creds = Credentials(api_key)  # credentials object to access GENAI

# Instantiate parameters for text generation
params = GenerateParams(decoding_method="sample", max_new_tokens=10)
prompts = ["Hello! How are you?", "How's the weather?"]

print(
    "==== Note that enum with symbols mapping to the same"
    " value does not yield to iterating over all the unique symbols,"
    " rather it represents all the symbols mapping to a single value"
    " by a single symbol during iteration. For example, if there are"
    " three symbols symb1, symb2, symb3 mapping to one value then"
    " during iteration it will do symb1 symb1 symb1 due to how it"
    " maps internally. ===="
)
for model_card in Model.models(credentials=creds):
    modelid = model_card.id
    model = Model(modelid, params=params, credentials=creds)
    responses = [response.generated_text for response in model.generate(prompts)]
    print(modelid, ":", responses)
    model = Model(modelid.value, params=params, credentials=creds)
    responses = [response.generated_text for response in model.generate(prompts)]
    print(modelid.value, ":", responses)



import os
import pathlib

from dotenv import load_dotenv

from genai.credentials import Credentials
from genai.model import Model
from genai.schemas import GenerateParams, ReturnOptions

# make sure you have a .env file under genai root with
# GENAI_KEY=<your-genai-key>
# GENAI_API=<genai-api-endpoint>
load_dotenv()
api_key = os.getenv("GENAI_KEY", None)
api_endpoint = os.getenv("GENAI_API", None)

print("\n------------- Example (Country-Capital-Factual-QA)-------------\n")

params = GenerateParams(
    decoding_method="sample",
    max_new_tokens=1,
    min_new_tokens=1,
    stream=False,
    temperature=0.7,
    top_k=50,
    top_p=1,
    return_options=ReturnOptions(input_text=False, input_tokens=True),
)

creds = Credentials(api_key, api_endpoint)
model = Model("google/flan-ul2", params=params, credentials=creds)

prompt_path = pathlib.Path(__file__, "..", "prompts", "Country-Capital-Factual-QA").resolve()
print(prompt_path)
# load a prompt from file
with open(prompt_path, "r") as f:
    prompt = f.read()

print(f"Prompt: \n {prompt}\n")

# Call generate function
responses = model.generate_as_completed([prompt])
for response in responses:
    print(f"Generated text: {response.generated_text}")



import pytest
from pydantic import BaseModel, ValidationError

from genai.utils.general import to_model_instance


@pytest.mark.unit
class TestUtils:
    def test_cast_to_model(self):
        class Company(BaseModel):
            name: str

        model = Company(name="IBM")
        assert model == to_model_instance(model, Company)
        assert model is not to_model_instance(model, Company)

        assert model == to_model_instance({"name": model.name}, Company)

        with pytest.raises(ValueError):
            to_model_instance([], Company)
        with pytest.raises(ValidationError):
            to_model_instance({}, Company)



import os

from dotenv import load_dotenv

from genai.credentials import Credentials
from genai.model import Model
from genai.schemas import GenerateParams, ReturnOptions

# make sure you have a .env file under genai root with
# GENAI_KEY=<your-genai-key>
# GENAI_API=<genai-api-endpoint>
load_dotenv()
api_key = os.getenv("GENAI_KEY", None)
api_endpoint = os.getenv("GENAI_API", None)
creds = Credentials(api_key=api_key, api_endpoint=api_endpoint)

print("\n------------- Example (Explain my code)-------------\n")

# Instantiate parameters for text generation
params = GenerateParams(
    decoding_method="sample", max_new_tokens=500, min_new_tokens=10, return_options=ReturnOptions(generated_tokens=True)
)

# Instantiate a model proxy object to send your requests
flan_ul2 = Model("google/flan-ul2", params=params, credentials=creds)

prompts = [" Explain life in one sentence"]
count = 0

for response in flan_ul2.generate_stream(prompts):
    print(response.input_token_count) if count == 0 else print(response.generated_text)
    count = count + 1



from genai.services.async_generator import AsyncResponseGenerator
from genai.services.request_handler import RequestHandler
from genai.services.service_interface import ServiceInterface

from genai.services.prompt_template_manager import PromptTemplateManager  # isort:skip
from genai.services.file_manager import FileManager  # isort:skip
from genai.services.tune_manager import TuneManager  # isort:skip

__all__ = [
    "RequestHandler",
    "ServiceInterface",
    "AsyncResponseGenerator",
    "PromptTemplateManager",
    "FileManager",
    "TuneManager",
]



import os

from dotenv import load_dotenv

from genai.services import ServiceInterface

# make sure you have a .env file under genai root with
# GENAI_KEY=<your-genai-key>
load_dotenv()

SERVICE_URL = "https://workbench-api.res.ibm.com"
GENAI_KEY = os.getenv("GENAI_KEY")
service = ServiceInterface(service_url=SERVICE_URL, api_key=GENAI_KEY)
completion = service.generate("google/ul2", ["hello! How are you?"])
print(completion)



from genai.schemas.files_params import FileListParams, MultipartFormData
from genai.services.request_handler import RequestHandler
from genai.utils.errors import to_genai_error
from genai.utils.request_utils import sanitize_params


class FilesRouter:
    FILES = "/v1/files"

    def __init__(self, service_url: str, api_key: str) -> None:
        self.service_url = service_url.rstrip("/")
        self.key = api_key

    def list_files(self, params: FileListParams = None):
        """List all files on the server.

        Args:
            params (FileListParams, optional): Parameters for listing files.

        Returns:
            Any: json from querying for file list.
        """
        try:
            params = sanitize_params(params)
            endpoint = self.service_url + FilesRouter.FILES
            return RequestHandler.get(endpoint, key=self.key, parameters=params)
        except Exception as e:
            raise to_genai_error(e)

    def get_file_metadata(self, file_id: str):
        """Get the file metadata from the server.

        Args:
            file_id (str): Id of the file to be retrieved.

        Returns:
            Any: json from querying for file retrieval.
        """
        try:
            endpoint = self.service_url + FilesRouter.FILES + "/" + file_id
            return RequestHandler.get(endpoint, key=self.key)
        except Exception as e:
            raise to_genai_error(e)

    def read_file(self, file_id: str):
        """Read the content of a file from the server.

        Args:
            file_id (str): Id of the file to be retrieved.

        Returns:
            Any: file contents.
        """
        try:
            endpoint = self.service_url + FilesRouter.FILES + "/" + file_id + "/content"
            return RequestHandler.get(endpoint, key=self.key)
        except Exception as e:
            raise to_genai_error(e)

    def delete_file(self, file_id: str):
        """Delete a file from the server.

        Args:
            file_id (str): Id of the file to be deleted.

        Returns:
            Any: json from querying for file deletion.
        """
        try:
            endpoint = self.service_url + FilesRouter.FILES + "/" + file_id
            return RequestHandler.delete(endpoint, key=self.key)
        except Exception as e:
            raise to_genai_error(e)

    def upload_file(self, multipart_form_data: MultipartFormData):
        """Upload a file to the server.

        Args:
            multipart_form_data (MultipartFormData): Form with the required params to be uploaded.

        Returns:
            Any: json from file uploaded.
        """
        try:
            endpoint = self.service_url + FilesRouter.FILES
            return RequestHandler.post(endpoint, key=self.key, files=dict(multipart_form_data))
        except Exception as e:
            raise to_genai_error(e)



import os

from dotenv import load_dotenv

try:
    from langchain.chains import LLMChain, SimpleSequentialChain
    from langchain.prompts import PromptTemplate
except ImportError:
    raise ImportError("Could not import langchain: Please install ibm-generative-ai[langchain] extension.")

from genai.credentials import Credentials
from genai.extensions.langchain import LangChainInterface
from genai.schemas import GenerateParams

# make sure you have a .env file under genai root with
# GENAI_KEY=<your-genai-key>
load_dotenv()
api_key = os.getenv("GENAI_KEY", None)
api_endpoint = os.getenv("GENAI_API", None)

params = GenerateParams(
    decoding_method="sample",
    max_new_tokens=100,
    min_new_tokens=1,
    stream=False,
    temperature=0.5,
    top_k=50,
    top_p=1,
).dict()  # Langchain uses dictionaries to pass kwargs

pt1 = PromptTemplate(
    input_variables=["topic"],
    template="Generate a random question about {topic}: Question: ",
)
pt2 = PromptTemplate(
    input_variables=["question"],
    template="Answer the following question: {question}",
)


creds = Credentials(api_key, api_endpoint)
flan = LangChainInterface(model="google/flan-ul2", credentials=creds, params=params)
model = LangChainInterface(model="google/flan-ul2", credentials=creds)
prompt_to_flan = LLMChain(llm=flan, prompt=pt1)
flan_to_model = LLMChain(llm=model, prompt=pt2)
qa = SimpleSequentialChain(chains=[prompt_to_flan, flan_to_model], verbose=True)
qa.run("life")



import os

from dotenv import load_dotenv

from genai.credentials import Credentials
from genai.model import Model
from genai.schemas import GenerateParams
from genai.schemas.generate_params import HAPOptions, ModerationsOptions

# make sure you have a .env file under genai root with
# GENAI_KEY=<your-genai-key>
# GENAI_API=<genai-api-endpoint>
load_dotenv()
api_key = os.getenv("GENAI_KEY", None)
api_endpoint = os.getenv("GENAI_API", None)
creds = Credentials(api_key, api_endpoint)  # credentials object to access GENAI

# Instantiate parameters for text generation
params = GenerateParams(
    decoding_method="sample",
    min_new_tokens=10,
    max_new_tokens=20,
    stream=True,
    moderations=ModerationsOptions(
        # Threshold is set to very low level to flag everything (testing purposes)
        # or set to True to enable HAP with default settings
        hap=HAPOptions(input=True, output=True, threshold=0.01)
    ),
)

# Instantiate a model proxy object to send your requests
flan_ul2 = Model("google/flan-ul2", params=params, credentials=creds)

prompts = ["Hello world!"]
for response in flan_ul2.generate_stream(prompts):
    print(f"Response: {response.generated_text}")
    print(f"Moderation: {response.moderation}")
    print("======")



import inspect
import os

from dotenv import load_dotenv

from genai.credentials import Credentials
from genai.model import Model
from genai.schemas import GenerateParams

# make sure you have a .env file under genai root with
# GENAI_KEY=<your-genai-key>
# GENAI_API=<genai-api-endpoint>
load_dotenv()
api_key = os.getenv("GENAI_KEY", None)
api_endpoint = os.getenv("GENAI_API", None)

print("\n------------- Example (Complete my code)-------------\n")

params = GenerateParams(
    decoding_method="sample",
    max_new_tokens=5,
    min_new_tokens=1,
    stream=False,
    temperature=0.7,
    top_k=50,
    top_p=1,
)

creds = Credentials(api_key, api_endpoint)
code_explainer = Model("google/flan-ul2", params=params, credentials=creds)


# pass in an actual python function to explain
def add_numbers(number_one, number_two):
    return number_one


prompt = inspect.getsource(add_numbers)
print(prompt + "\n")
responses = code_explainer.generate([prompt])
for response in responses:
    print(f"Generated text:\n {response.generated_text}")



import pytest

from genai import Credentials


@pytest.mark.unit
class TestCredentials:
    @pytest.mark.filterwarnings("ignore::DeprecationWarning")
    def test_remove_version(self):
        credentials = Credentials(api_key="GENAI_API_KEY", api_endpoint="https://workbench-api.res.ibm.com/v1/")

        assert "/v1" not in credentials.api_endpoint

        credentials = Credentials(api_key="GENAI_API_KEY", api_endpoint="https://workbench-api.res.ibm.com/v2/ai/v12")

        assert credentials.api_endpoint == "https://workbench-api.res.ibm.com/v2/ai"

    def test_trigger_deprecation_warning(self):
        with pytest.deprecated_call():
            Credentials(api_key="GENAI_API_KEY", api_endpoint="https://workbench-api.res.ibm.com/v1/")



import cachetools.func
from httpx import ConnectError, Response

from genai.exceptions import GenAiException
from genai.options import Options
from genai.routers import FilesRouter, PromptTemplateRouter, TunesRouter
from genai.schemas import GenerateParams, HistoryParams, TokenParams
from genai.schemas.responses import GenerateLimits
from genai.services.request_handler import RequestHandler
from genai.utils.errors import to_genai_error
from genai.utils.request_utils import sanitize_params

__all__ = ["ServiceInterface"]


class ServiceInterface:
    GENERATE = "/v1/generate"
    TOKENIZE = "/v1/tokenize"
    HISTORY = "/v1/requests"
    TOU = "/v1/user"
    MODELS = "/v1/models"
    GENERATE_LIMITS = "/v1/generate/limits"

    def __init__(self, service_url: str, api_key: str) -> None:
        """Initialize ServiceInterface.

        Args:
            service_url (str): Base URL for querying.
            api_key (str): User API key for authorization.
        """
        self.service_url = service_url.rstrip("/")
        self.key = api_key
        self._prompt_templating = PromptTemplateRouter(service_url=service_url, api_key=api_key)
        self._files = FilesRouter(service_url=service_url, api_key=api_key)
        self._tunes = TunesRouter(service_url=service_url, api_key=api_key)

    def models(self):
        """Generate a completion text for the given model, inputs, and params.

        Returns:
            Any: json from querying for text completion.
        """
        try:
            endpoint = self.service_url + ServiceInterface.MODELS
            return RequestHandler.get(endpoint, key=self.key)
        except ConnectError as e:
            raise GenAiException(
                Exception("Endpoint unreachable. Please check connectivity.\nRaw error message = {}".format(e))
            )
        except Exception as e:
            raise to_genai_error(e)

    @cachetools.func.ttl_cache(ttl=1000, typed=True, maxsize=None)
    def generate_limits(self):
        """Retrieve generate limits for the account associated with current auth token."""
        try:
            endpoint = f"{self.service_url}{ServiceInterface.GENERATE_LIMITS}"
            response = RequestHandler.get(endpoint, key=self.key)
            if not response.is_success:
                raise GenAiException(response)
            return GenerateLimits(**response.json())
        except ConnectError as e:
            raise GenAiException("Endpoint unreachable. Please check connectivity.\nRaw error message = {}".format(e))
        except Exception as e:
            raise to_genai_error(e)

    def generate(
        self,
        model: str,
        inputs: list,
        params: GenerateParams = None,
        streaming: bool = False,
        options: Options = None,
    ):
        """Generate a completion text for the given model, inputs, and params.

        Args:
            model (str): Model id.
            inputs (list): List of inputs.
            params (GenerateParams, optional): Parameters for generation. Defaults to None.
            streaming (bool, optional): Streaming response flag. Defaults to False.
            options (Options, optional): Additional parameters to pass in the query payload. Defaults to None.

        Returns:
            Any: json from querying for text completion.
        """
        try:
            params = sanitize_params(params)
            endpoint = self.service_url + ServiceInterface.GENERATE
            res = RequestHandler.post(
                endpoint,
                key=self.key,
                model_id=model,
                inputs=inputs,
                parameters=params,
                streaming=streaming,
                options=options,
            )
            return res
        except Exception as e:
            raise to_genai_error(e)

    def tokenize(
        self,
        model: str,
        inputs: list,
        params: TokenParams = None,
        options: Options = None,
    ):
        """Do the conversion of provided inputs to tokens for a given model.

        Args:
            model (str): Model id.
            inputs (list): List of inputs.
            params (TokenParams, optional): Parameters for generation. Defaults to None.
            options (Options, optional): Additional parameters to pass in the query payload. Defaults to None.

        Returns:
            Any: json from querying for tokenization.
        """
        try:
            params = sanitize_params(params)
            endpoint = self.service_url + ServiceInterface.TOKENIZE
            return RequestHandler.post(
                endpoint,
                key=self.key,
                model_id=model,
                inputs=inputs,
                parameters=params,
                options=options,
            )
        except Exception as e:
            raise to_genai_error(e)

    def history(self, params: HistoryParams = None):
        """Retrieve past generation requests and responses returned by the given models.

        Args:
            params (HistoryParams, optional): Parameters for retrieving history. Defaults to None.

        Returns:
            Any: json from querying for tokenization.
        """
        try:
            params = sanitize_params(params)
            endpoint = self.service_url + ServiceInterface.HISTORY
            return RequestHandler.get(endpoint, key=self.key, parameters=params)
        except Exception as e:
            raise to_genai_error(e)

    def terms_of_use(self, accept: bool) -> Response:
        """Accept the API Terms of Use

        Args:
            accept (bool): If the user accepts the TOU

        Raises:
            GenAiException: A general GenAI exception if there is an issue
                with the request

        Returns:
            httpx.Response: Response from REST API
        """
        tou_payload = {"tou_accepted": accept}

        try:
            endpoint = self.service_url + ServiceInterface.TOU
            return RequestHandler.patch(endpoint, key=self.key, json_data=tou_payload)
        except Exception as e:
            raise to_genai_error(e)

    # * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    #   ASYNC
    # * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    async def async_generate(self, model, inputs, params: GenerateParams = None, options: Options = None):
        """Generate a completion text for the given model, inputs, and params.

        Args:
            model (str): Model id.
            inputs (list): List of inputs.
            params (GenerateParams, optional): Parameters for generation. Defaults to None.
            options (Options, optional): Additional parameters to pass in the query payload. Defaults to None.

        Returns:
            Any: json from querying for text completion.
        """
        try:
            params = sanitize_params(params)
            endpoint = self.service_url + ServiceInterface.GENERATE
            return await RequestHandler.async_generate(
                endpoint,
                key=self.key,
                model_id=model,
                inputs=inputs,
                parameters=params,
                options=options,
            )
        except Exception as e:
            # without VPN this will fail
            raise to_genai_error(e)

    async def async_tokenize(self, model, inputs, params: TokenParams = None, options: Options = None):
        """Do the conversion of provided inputs to tokens for a given model.

        Args:
            model (str): Model id.
            inputs (list): List of inputs.
            params (TokenParams, optional): Parameters for generation. Defaults to None.
            options (Options, optional): Additional parameters to pass in the query payload. Defaults to None.

        Returns:
            Any: json from querying for tokenization.
        """
        try:
            params = sanitize_params(params)
            endpoint = self.service_url + ServiceInterface.TOKENIZE
            return await RequestHandler.async_tokenize(
                endpoint,
                key=self.key,
                model_id=model,
                inputs=inputs,
                parameters=params,
                options=options,
            )
        except Exception as e:
            raise to_genai_error(e)

    async def async_history(self, params: HistoryParams = None):
        """Retrieve past generation requests and responses returned by the given models.

        Args:
            params (HistoryParams, optional): Parameters for retrieving history. Defaults to None.

        Returns:
            Any: json from querying for tokenization.
        """
        try:
            params = sanitize_params(params)
            endpoint = self.service_url + ServiceInterface.HISTORY
            return await RequestHandler.async_get(endpoint, key=self.key, parameters=params)
        except Exception as e:
            raise to_genai_error(e)

    async def async_terms_of_use(self, accept: bool) -> Response:
        """Accept the API Terms of Use

        Args:
            accept (bool): If the user accepts the TOU

        Raises:
            GenAiException: A general GenAI exception if there is an issue
                with the request

        Returns:
            httpx.Response: Response from REST API
        """
        tou_payload = {"tou_accepted": accept}

        try:
            endpoint = self.service_url + ServiceInterface.TOU
            return await RequestHandler.async_patch(endpoint, key=self.key, json_data=tou_payload)
        except Exception as e:
            raise to_genai_error(e)



import json
import logging
import queue
import random
import re
import time
from unittest.mock import AsyncMock, patch

import httpx
import pytest
from pytest_httpx import HTTPXMock

from genai.exceptions import GenAiException
from genai.schemas import GenerateParams, ReturnOptions, TokenParams
from genai.schemas.responses import GenerateResponse, TokenizeResponse
from genai.services import AsyncResponseGenerator, RequestHandler, ServiceInterface
from genai.services.connection_manager import ConnectionManager
from genai.utils.request_utils import sanitize_params
from tests.assets.response_helper import SimpleResponse

logger = logging.getLogger(__name__)


@pytest.mark.unit
class TestAsyncResponseGenerator:
    def setup_method(self):
        self.service = ServiceInterface(service_url="http://SERVICE_URL", api_key="API_KEY")
        self.model = "google/ul2"
        self.inputs = ["Write a tagline for an alumni association: Together we"]

    @pytest.fixture
    def mock_generate_json(self, mocker):
        async_mock = AsyncMock()
        mocker.patch(
            "genai.services.AsyncResponseGenerator._get_response_json",
            side_effect=async_mock,
        )
        return async_mock

    @pytest.fixture
    def generate_params(self):
        return GenerateParams(
            temperature=0.05,
            max_new_tokens=3,
            return_options=ReturnOptions(input_text=True),
        )

    @pytest.fixture
    def mock_tokenize_json(self, mocker):
        async_mock = AsyncMock()
        mocker.patch(
            "genai.services.AsyncResponseGenerator._get_response_json",
            side_effect=async_mock,
        )
        return async_mock

    @pytest.fixture
    def tokenize_params(self):
        return TokenParams(return_tokens=True)

    @pytest.mark.asyncio
    async def test_concurrent_generate(self, mock_generate_json, generate_params):
        expected = SimpleResponse.generate(model=self.model, inputs=self.inputs)
        mock_generate_json.return_value = expected
        num_prompts = 3

        counter = 0
        with AsyncResponseGenerator(
            self.model, self.inputs * num_prompts, generate_params, self.service
        ) as asynchelper:
            assert ConnectionManager.async_generate_client is not None
            for result in asynchelper.generate_response():
                assert result == GenerateResponse(**expected).results[0]
                counter += 1
            assert counter == num_prompts
        assert ConnectionManager.async_generate_client is None

    @pytest.mark.asyncio
    async def test_concurrent_tokenize(self, mock_tokenize_json, tokenize_params):
        expected = SimpleResponse.tokenize(model=self.model, inputs=self.inputs, params=tokenize_params)
        mock_tokenize_json.return_value = expected

        with AsyncResponseGenerator(
            self.model, self.inputs, tokenize_params, self.service, fn="tokenize"
        ) as asynchelper:
            assert ConnectionManager.async_tokenize_client is not None
            for result in asynchelper.generate_response():
                assert result == TokenizeResponse(**expected).results[0]
        assert ConnectionManager.async_tokenize_client is None

    @pytest.mark.asyncio
    async def test_async_helper_total_generate_responses(self, mock_generate_json, generate_params):
        num_prompts = 7
        single_response = SimpleResponse.generate(model=self.model, inputs=self.inputs, params=generate_params)

        def side_effect(model, inputs, params):
            response = {}
            response["model_id"] = single_response["model_id"]
            response["created_at"] = single_response["created_at"]
            response["results"] = single_response["results"] * len(inputs)
            return response

        mock_generate_json.side_effect = side_effect

        with AsyncResponseGenerator(
            self.model, self.inputs * num_prompts, generate_params, self.service
        ) as asynchelper:
            assert ConnectionManager.async_generate_client is not None
            total_responses = 0
            for _ in asynchelper.generate_response():
                total_responses += 1
            assert total_responses == num_prompts
        assert ConnectionManager.async_generate_client is None

    @pytest.mark.asyncio
    async def test_async_helper_total_tokenize_responses(self, mock_tokenize_json, tokenize_params):
        num_prompts = 7
        single_response = SimpleResponse.tokenize(model=self.model, inputs=self.inputs, params=tokenize_params)

        def side_effect(model, inputs, params):
            response = {}
            response["model_id"] = single_response["model_id"]
            response["created_at"] = single_response["created_at"]
            response["results"] = single_response["results"] * len(inputs)
            return response

        mock_tokenize_json.side_effect = side_effect

        with AsyncResponseGenerator(
            self.model,
            self.inputs * num_prompts,
            tokenize_params,
            self.service,
            fn="tokenize",
        ) as asynchelper:
            assert ConnectionManager.async_tokenize_client is not None
            total_responses = 0
            for _ in asynchelper.generate_response():
                total_responses += 1
            assert total_responses == num_prompts
        assert ConnectionManager.async_tokenize_client is None

    @pytest.mark.asyncio
    async def test_concurrent_tokenize_dropped_request(self, httpx_mock, tokenize_params):
        # Return a 429 for a tokenize request from RequestHandler
        httpx_mock.add_response(method="POST", status_code=429, json={})
        num_prompts = 17
        counter = 0
        with AsyncResponseGenerator(
            self.model,
            self.inputs * num_prompts,
            tokenize_params,
            self.service,
            fn="tokenize",
        ) as asynchelper:
            for result in asynchelper.generate_response():
                assert result is None
                counter += 1
            assert counter == num_prompts

    @pytest.mark.asyncio
    @pytest.mark.skip(reason="pytest_httpx does not handle custom transports")
    async def test_concurrent_generate_retry(self, httpx_mock: HTTPXMock, generate_params):
        with patch.multiple(AsyncResponseGenerator, MAX_RETRIES_GENERATE=1):
            for code in [httpx.codes.SERVICE_UNAVAILABLE, httpx.codes.TOO_MANY_REQUESTS]:
                httpx_mock.add_response(method="POST", status_code=code, json={})
                httpx_mock.add_response(method="POST", status_code=httpx.codes.OK, json={})

                with AsyncResponseGenerator(self.model, self.inputs, generate_params, self.service) as asynchelper:
                    response = list(asynchelper.generate_response())
                    assert len(response) == 2
                    assert response[0] is None
                    assert response[1] is not None

                requests = httpx_mock.get_requests(url=re.compile(f".+{ServiceInterface.GENERATE}$"), method="POST")
                assert len(requests) == 2

    @pytest.mark.asyncio
    @pytest.mark.skip(reason="pytest_httpx does not handle custom transports")
    async def test_concurrent_tokenize_retry(self, httpx_mock, tokenize_params):
        saved = ConnectionManager.MAX_RETRIES_TOKENIZE
        ConnectionManager.MAX_RETRIES_TOKENIZE = 2
        for code in [httpx.codes.SERVICE_UNAVAILABLE, httpx.codes.TOO_MANY_REQUESTS]:
            httpx_mock.add_response(method="POST", status_code=code, json={})
            with AsyncResponseGenerator(
                self.model, self.inputs, tokenize_params, self.service, fn="tokenize"
            ) as asynchelper:
                time_start = time.time()
                for result in asynchelper.generate_response():
                    assert result is None
                time_end = time.time()
                assert time_end - time_start > 5
        ConnectionManager.MAX_RETRIES_TOKENIZE = saved

    @pytest.mark.asyncio
    async def test_concurrent_generate_dropped_request(self, httpx_mock, generate_params):
        failed_id = 4
        single_response = SimpleResponse.generate(model=self.model, inputs=self.inputs, params=generate_params)
        headers, json_data, _ = RequestHandler._metadata(
            "POST",
            key="TEST_KEY",
            model_id=self.model,
            inputs=["Input " + str(failed_id)],
            parameters=sanitize_params(generate_params),
        )
        # Following two lines: Selected input id should return 429, others should succeed
        httpx_mock.add_response(method="POST", json=single_response)
        httpx_mock.add_response(
            method="POST",
            status_code=429,
            match_content=bytes(json.dumps(json_data), encoding="utf-8"),
        )
        num_prompts = 9
        counter = 0
        inputs = ["Input " + str(i) for i in range(num_prompts)]
        with AsyncResponseGenerator(self.model, inputs, generate_params, self.service) as asynchelper:
            assert ConnectionManager.async_generate_client is not None
            for _ in asynchelper.generate_response():
                if counter == failed_id:
                    # print(response)
                    assert True
                    # assert response is None
                else:
                    assert True
                    # assert response is not None
                counter += 1
            assert counter == num_prompts
        assert ConnectionManager.async_generate_client is None

    @pytest.mark.asyncio
    async def test_concurrent_generate_inorder(self, mock_generate_json, generate_params, mocker):
        num_prompts = 31
        inputs = ["This is input number " + str(i) for i in range(num_prompts)]
        expected = SimpleResponse.generate_response_array_async(model=self.model, inputs=inputs)
        mock_generate_json.side_effect = expected

        counter = 0
        permutation = list(range(num_prompts))
        random.shuffle(permutation)
        # Permute what comes out of the queue completely but expect results in order
        with AsyncResponseGenerator(self.model, inputs, generate_params, self.service, ordered=True) as asynchelper:
            assert ConnectionManager.async_generate_client is not None
            mocker.patch.object(
                queue.Queue,
                "get",
                spec=queue.Queue,
                side_effect=[
                    (
                        permutation[counter],
                        1,
                        GenerateResponse(**expected[permutation[counter]]),
                        None,
                    )
                    for counter in range(num_prompts)
                ],
            )
            mocker.patch.object(queue.Queue, "put_nowait", spec=queue.Queue, return_value=None)
            mocker.patch.object(queue.Queue, "task_done", spec=queue.Queue, return_value=None)
            for result in asynchelper.generate_response():
                expected_result = GenerateResponse(**expected[counter]).results[0]
                assert result.generated_text == expected_result.generated_text
                assert result.input_text == inputs[counter]
                counter += 1
                mock_generate_json.return_value = expected[counter] if counter < num_prompts else None
            assert counter == num_prompts
        assert ConnectionManager.async_generate_client is None

    @pytest.mark.asyncio
    async def test_concurrent_generate_not_inorder(self, mock_generate_json, generate_params, mocker):
        num_prompts = 5
        inputs = ["This is input number " + str(i) for i in range(num_prompts)]
        expected = SimpleResponse.generate_response_array_async(model=self.model, inputs=inputs)
        mock_generate_json.side_effect = expected

        counter = 0
        permutation = list(range(num_prompts))
        random.shuffle(permutation)

        # Permute what comes out of the queue completely and expect results in the same order
        with AsyncResponseGenerator(self.model, inputs, generate_params, self.service, ordered=False) as asynchelper:
            assert ConnectionManager.async_generate_client is not None
            mocker.patch.object(
                queue.Queue,
                "get",
                spec=queue.Queue,
                side_effect=[
                    (
                        permutation[counter],
                        1,
                        GenerateResponse(**expected[permutation[counter]]),
                        None,
                    )
                    for counter in range(num_prompts)
                ],
            )
            mocker.patch.object(queue.Queue, "put_nowait", spec=queue.Queue, return_value=None)
            mocker.patch.object(queue.Queue, "task_done", spec=queue.Queue, return_value=None)
            for result in asynchelper.generate_response():
                expected_result = GenerateResponse(**expected[permutation[counter]]).results[0]
                assert result.generated_text == expected_result.generated_text
                assert result.input_text == inputs[permutation[counter]]
                counter += 1
                mock_generate_json.return_value = expected[counter] if counter < num_prompts else None
            assert counter == num_prompts
        assert ConnectionManager.async_generate_client is None

    @pytest.mark.asyncio
    async def test_concurrent_tokenize_inorder(self, mock_tokenize_json, tokenize_params, mocker):
        num_prompts = 143
        inputs = ["This is input number " + str(i) for i in range(num_prompts)]
        expected = SimpleResponse.tokenize_response_array_async(model=self.model, inputs=inputs)
        mock_tokenize_json.side_effect = expected

        permutation = list(range(len(expected)))
        random.shuffle(permutation)
        # Permute what comes out of the queue completely but expect results in order
        with AsyncResponseGenerator(
            self.model,
            inputs,
            tokenize_params,
            self.service,
            fn="tokenize",
            ordered=True,
        ) as asynchelper:
            assert ConnectionManager.async_tokenize_client is not None
            mocker.patch.object(
                queue.Queue,
                "get",
                spec=queue.Queue,
                side_effect=[
                    (
                        permutation[counter],
                        asynchelper._batch_size,
                        TokenizeResponse(**expected[permutation[counter]]),
                        None,
                    )
                    for counter in range(len(expected))
                ],
            )
            mocker.patch.object(queue.Queue, "put_nowait", spec=queue.Queue, return_value=None)
            mocker.patch.object(queue.Queue, "task_done", spec=queue.Queue, return_value=None)
            num_results = 0
            observed = list(asynchelper.generate_response())
            for counter in range(asynchelper._num_batches):
                response = TokenizeResponse(**expected[counter])
                for result in response.results:
                    assert observed[num_results].tokens == result.tokens
                    assert observed[num_results].input_text == result.input_text
                    num_results += 1
            assert num_prompts == num_results
        assert ConnectionManager.async_tokenize_client is None

    @pytest.mark.asyncio
    async def test_concurrent_tokenize_not_inorder(self, mock_tokenize_json, tokenize_params, mocker):
        num_prompts = 143
        inputs = ["This is input number " + str(i) for i in range(num_prompts)]
        expected = SimpleResponse.tokenize_response_array_async(model=self.model, inputs=inputs)
        mock_tokenize_json.side_effect = expected

        permutation = list(range(len(expected)))
        random.shuffle(permutation)
        # Permute what comes out of the queue completely and expect results in the same order
        with AsyncResponseGenerator(
            self.model,
            inputs,
            tokenize_params,
            self.service,
            fn="tokenize",
            ordered=False,
        ) as asynchelper:
            assert ConnectionManager.async_tokenize_client is not None
            mocker.patch.object(
                queue.Queue,
                "get",
                spec=queue.Queue,
                side_effect=[
                    (
                        permutation[counter],
                        asynchelper._batch_size,
                        TokenizeResponse(**expected[permutation[counter]]),
                        None,
                    )
                    for counter in range(len(expected))
                ],
            )
            mocker.patch.object(queue.Queue, "put_nowait", spec=queue.Queue, return_value=None)
            mocker.patch.object(queue.Queue, "task_done", spec=queue.Queue, return_value=None)
            num_results = 0
            observed = list(asynchelper.generate_response())
            for counter in range(asynchelper._num_batches):
                response = TokenizeResponse(**expected[permutation[counter]])
                for result in response.results:
                    assert observed[num_results].tokens == result.tokens
                    assert observed[num_results].input_text == result.input_text
                    num_results += 1
            assert num_prompts == num_results
        assert ConnectionManager.async_tokenize_client is None

    @pytest.mark.asyncio
    async def test_concurrent_tokenize_nones(self, mock_tokenize_json, tokenize_params, mocker):
        # test that if one request gets dropped, we get appropriate number of nones
        num_prompts = 18
        inputs = ["This is input number " + str(i) for i in range(num_prompts)]
        expected = SimpleResponse.tokenize_response_array_async(model=self.model, inputs=inputs)
        mock_tokenize_json.side_effect = expected

        batch_indices = list(range(len(expected)))
        side_effect = [
            (
                batch_indices[counter],
                5,
                TokenizeResponse(**expected[batch_indices[counter]]),
                None,
            )
            for counter in range(len(expected))
        ]
        # Test for failing requests for the first batch, last batch, intermediate batch
        for failed_id in [0, -1, random.randint(1, len(expected) - 2)]:
            batch_size = num_prompts % 5 if failed_id == -1 else 5
            side_effect_with_failed = side_effect[:]
            side_effect_with_failed[failed_id] = (
                batch_indices[failed_id],
                batch_size,
                None,
                GenAiException("Something went wrong!"),
            )

            # Permute what comes out of the queue completely and expect results in the same order
            with AsyncResponseGenerator(
                self.model,
                inputs,
                tokenize_params,
                self.service,
                fn="tokenize",
                ordered=False,
            ) as asynchelper:
                assert ConnectionManager.async_tokenize_client is not None
                mocker.patch.object(
                    queue.Queue,
                    "get",
                    spec=queue.Queue,
                    side_effect=side_effect_with_failed,
                )
                mocker.patch.object(queue.Queue, "put_nowait", spec=queue.Queue, return_value=None)
                mocker.patch.object(queue.Queue, "task_done", spec=queue.Queue, return_value=None)
                num_results = 0
                observed = list(asynchelper.generate_response())
                for counter in range(asynchelper._num_batches):
                    response = TokenizeResponse(**expected[counter])
                    for result in response.results:
                        if failed_id == 0 and num_results < 5:
                            assert observed[num_results] is None
                        elif failed_id == -1 and num_results >= num_prompts - num_prompts % 5:
                            assert observed[num_results] is None
                        elif (
                            failed_id * asynchelper._batch_size
                            <= num_results
                            < (failed_id + 1) * asynchelper._batch_size
                        ):
                            assert observed[num_results] is None
                        else:
                            assert observed[num_results].tokens == result.tokens
                            assert observed[num_results].input_text == result.input_text
                        num_results += 1
                assert num_prompts == num_results
            assert ConnectionManager.async_tokenize_client is None



import os

from dotenv import load_dotenv

from genai.credentials import Credentials
from genai.model import Model

# make sure you have a .env file under genai root with
# GENAI_KEY=<your-genai-key>
# GENAI_API=<your-genai-api endpoint>
load_dotenv()
api_key = os.getenv("GENAI_KEY", None)
api_url = os.getenv("GENAI_API", None)
creds = Credentials(api_key, api_endpoint=api_url)

print("======= List of all models =======")
for m in Model.models(credentials=creds):
    print(m)

print("====== Checking model availability =======")
model = Model("google/ul2", params=None, credentials=creds)
print("Model availability for 'google/ul2': ", model.available())

model = Model("random", params=None, credentials=creds)
print("Model availability for 'random': ", model.available())

print("====== Display model card =======")
model = Model("google/ul2", params=None, credentials=creds)
print("Model info for 'google/ul2': \n", model.info())



import logging
import os

from dotenv import load_dotenv

from genai.model import Credentials, Model
from genai.schemas import GenerateParams
from genai.utils.http_provider import HttpProvider

logging.basicConfig(level="ERROR")

# Modify options for `httpx` library which the SDK uses
HttpProvider.default_http_client_options["verify"] = False
HttpProvider.default_http_transport_options["verify"] = False

# make sure you have a .env file under genai root with
# GENAI_KEY=<your-genai-key>
load_dotenv()
api_key = os.getenv("GENAI_KEY", None)
api_url = os.getenv("GENAI_API", None)
creds = Credentials(api_key=api_key, api_endpoint=api_url)


# Instantiate parameters for text generation
generate_params = GenerateParams(decoding_method="sample", max_new_tokens=5, min_new_tokens=1)
flan_ul2 = Model("google/flan-ul2", params=generate_params, credentials=creds)
[response] = flan_ul2.generate(prompts=["Generate a random number"])
print(f"Your random number is: {response.generated_text}")



import json
import logging
import time
from collections import deque
from collections.abc import Generator
from typing import Any, Callable, List, Optional, Union

from tqdm import tqdm

from genai.credentials import Credentials
from genai.exceptions import GenAiException
from genai.metadata import Metadata
from genai.options import Options
from genai.prompt_pattern import PromptPattern
from genai.schemas import GenerateParams, TokenParams
from genai.schemas.responses import (
    GenerateResponse,
    GenerateResult,
    GenerateStreamResponse,
    ModelCard,
    ModelList,
    TokenizeResponse,
    TokenizeResult,
)
from genai.schemas.tunes_params import (
    CreateTuneHyperParams,
    CreateTuneParams,
    DownloadAssetsParams,
    TunesListParams,
)
from genai.services import AsyncResponseGenerator, ServiceInterface
from genai.services.tune_manager import TuneManager
from genai.utils.errors import to_genai_error
from genai.utils.general import to_model_instance
from genai.utils.service_utils import _get_service

logger = logging.getLogger(__name__)


class Model:
    _accessors = set()

    def __init__(
        self,
        model: str,
        params: Union[GenerateParams, TokenParams, Any] = None,
        credentials: Credentials = None,
    ):
        """Instantiates the Model Interface

        Args:
            model (str): The type of model to use
            params (Union[GenerateParams, TokenParams]): Parameters to use during generate requests
            credentials (Credentials): The API Credentials
        """
        logger.debug(f"Model Created:  Model: {model}, endpoint: {credentials.api_endpoint}")
        self.model = model
        self.params = params
        self.creds = credentials
        self.service = ServiceInterface(service_url=credentials.api_endpoint, api_key=credentials.api_key)

    def generate_stream(
        self, prompts: Union[list[str], list[PromptPattern]], options: Options = None
    ) -> Generator[GenerateStreamResponse]:
        if len(prompts) > 0 and isinstance(prompts[0], PromptPattern):
            prompts = PromptPattern.list_str(prompts)

        params = to_model_instance(self.params, GenerateParams)
        params.stream = True

        try:
            for i in range(0, len(prompts), Metadata.DEFAULT_MAX_PROMPTS):
                batch = prompts[i : min(i + Metadata.DEFAULT_MAX_PROMPTS, len(prompts))]

                response_gen = self.service.generate(self.model, batch, params, options=options, streaming=True)

                for response in response_gen:
                    if not response:
                        continue

                    if "status_code" in response:
                        error = json.loads(response)
                        raise GenAiException(error)

                    try:
                        parsed_response = json.loads(response)
                        if "moderation" in parsed_response:
                            result = {
                                "id": parsed_response["id"],
                                "results": [],
                                "model_id": parsed_response["model_id"],
                                "created_at": parsed_response["created_at"],
                                "moderation": parsed_response["moderation"],
                            }
                            yield GenerateStreamResponse(**result)

                        for result in parsed_response["results"]:
                            yield GenerateStreamResponse(**result)

                    except Exception:
                        logger.error("Could not parse {} as literal_eval".format(response))

        except Exception as ex:
            raise to_genai_error(ex)

    def generate_as_completed(
        self, prompts: Union[list[str], list[PromptPattern]], options: Options = None
    ) -> Generator[GenerateResult]:
        """The generate endpoint is the centerpiece of the GENAI alpha.
        It provides a simplified and flexible, yet powerful interface to the supported
        models as a service. Given a text prompt as inputs, and required parameters
        the selected model (model_id) will generate a completion text as generated_text.

        Args:
            prompts (list[str]): The list of one or more prompt strings.
            options (Options, optional): Additional parameters to pass in the query payload. Defaults to None.

        Yields:
            Generator[GenerateResult]: A generator of results
        """
        if len(prompts) > 0 and isinstance(prompts[0], PromptPattern):
            prompts = PromptPattern.list_str(prompts)

        params = to_model_instance(self.params, GenerateParams)
        params.stream = False

        logger.debug(f"Calling Generate. Prompts: {prompts}, params: {params}")

        def get_remaining_limit():
            limits = self.service.generate_limits()
            return limits.tokenCapacity - limits.tokensUsed

        def execute(inputs: List[str], attempt=0) -> List[GenerateResult]:
            response = self.service.generate(
                model=self.model,
                inputs=inputs,
                params=params,
                options=options,
            )
            if response.status_code == 200:
                raw_response = response.json()
                for i, result in enumerate(raw_response["results"]):
                    result["input_text"] = inputs[i]
                generate_response = GenerateResponse(**raw_response)
                return generate_response.results
            elif response.status_code == 429:
                nonlocal remaining_limit
                remaining_limit = 0
                time.sleep(2 ** (attempt + 1))
                return execute(inputs, attempt + 1)
            else:
                raise GenAiException(response)

        try:
            remaining_prompts = deque(prompts)
            remaining_limit = get_remaining_limit()
            while remaining_prompts:
                while remaining_limit <= 0:
                    time.sleep(1)
                    remaining_limit = get_remaining_limit()

                prompts_to_process = min(remaining_limit, Metadata.DEFAULT_MAX_PROMPTS, len(remaining_prompts))
                remaining_limit -= prompts_to_process
                inputs = [remaining_prompts.popleft() for _ in range(prompts_to_process)]
                for result in execute(inputs):
                    yield result
        except Exception as ex:
            raise to_genai_error(ex)

    def generate(self, prompts: Union[list[str], list[PromptPattern]], options: Options = None) -> list[GenerateResult]:
        """The generate endpoint is the centerpiece of the GENAI alpha.
        It provides a simplified and flexible, yet powerful interface to the supported
        models as a service. Given a text prompt as inputs, and required parameters
        the selected model (model_id) will generate a completion text as generated_text.

        Args:
            prompts (list[str]): The list of one or more prompt strings.
            options (Options, optional): Additional parameters to pass in the query payload. Defaults to None.

        Returns:
            list[GenerateResult]: A list of results
        """
        return list(self.generate_as_completed(prompts, options))

    def generate_async(
        self,
        prompts: Union[list[str], list[PromptPattern]],
        ordered: bool = False,
        callback: Optional[Callable[[GenerateResult], Any]] = None,
        hide_progressbar: bool = False,
        options: Optional[Options] = None,
        *,
        throw_on_error: bool = False,
        max_concurrency_limit: Optional[int] = None,
    ) -> Generator[Union[GenerateResult, None], None, None]:
        """The generate endpoint is the centerpiece of the GENAI alpha.
        It provides a simplified and flexible, yet powerful interface to the supported
        models as a service. Given a text prompt as inputs, and required parameters
        the selected model (model_id) will generate a completion text as generated_text.
        This python method generates responses utilizing async capabilities and returns
        responses as they arrive.

        Args:
            prompts (list[str]): The list of one or more prompt strings.
            ordered (bool): Whether the responses should be returned in-order.
            callback (Callable[[GenerateResult], Any]): Optional callback
                to be called after generating result for a prompt.
            hide_progressbar (bool, optional): boolean flag to hide or show a progress bar.
            options (Options, optional): Additional parameters to pass in the query payload. Defaults to None.
            throw_on_error (bool, optional): Throws error on failure. Defaults to False.
            max_concurrency_limit (int, optional): Maximum number of concurrent requests to make. Defaults to None.
        Returns:
            Generator[Union[GenerateResult, None]]: A list of results
        """
        if len(prompts) > 0 and isinstance(prompts[0], PromptPattern):
            prompts = PromptPattern.list_str(prompts)

        params = to_model_instance(self.params, GenerateParams)
        params.stream = False
        logger.debug(f"Calling Generate. Prompts: {prompts}, Params: {params}")

        try:
            with AsyncResponseGenerator(
                model_id=self.model,
                prompts=prompts,
                params=params,
                service=self.service,
                ordered=ordered,
                callback=callback,
                options=options,
                throw_on_error=throw_on_error,
                max_concurrency_limit=max_concurrency_limit,
            ) as generator:
                for response in tqdm(
                    iterable=generator.generate_response(),
                    total=len(prompts),
                    desc="Progress",
                    unit=" inputs",
                    disable=hide_progressbar,
                ):
                    yield response
        except Exception as ex:
            raise to_genai_error(ex)

    def tokenize_as_completed(
        self,
        prompts: Union[list[str], list[PromptPattern]],
        return_tokens: bool = False,
        options: Options = None,
    ) -> Generator[TokenizeResult, None, None]:
        """The tokenize endpoint allows you to check the conversion of provided prompts to tokens
        for a given model. It splits text into words or subwords, which then are converted to ids
        through a look-up table (vocabulary). Tokenization allows the model to have a reasonable
        vocabulary size.

        Args:
            prompts (Union[list[str], list[PromptPattern]]): Prompts to be tokenized.
            return_tokens (bool, optional): Return tokens with the response. Defaults to False.
            options (Options, optional): Additional parameters to pass in the query payload. Defaults to None.
        Yields:
            Generator[TokenizeResult]: The Tokenized input
        """
        if len(prompts) > 0 and isinstance(prompts[0], PromptPattern):
            prompts = PromptPattern.list_str(prompts)

        try:
            params = TokenParams(return_tokens=return_tokens)
            for i in range(0, len(prompts), Metadata.DEFAULT_MAX_PROMPTS):
                tokenize_response = self.service.tokenize(
                    model=self.model,
                    inputs=prompts[i : min(i + Metadata.DEFAULT_MAX_PROMPTS, len(prompts))],
                    params=params,
                    options=options,
                )

                if tokenize_response.status_code == 200:
                    response_json = tokenize_response.json()
                    for y, result in enumerate(response_json["results"]):
                        result["input_text"] = prompts[i + y]
                    responses = TokenizeResponse(**response_json)
                    for token in responses.results:
                        yield token
                else:
                    raise GenAiException(tokenize_response)
        except Exception as ex:
            raise to_genai_error(ex)

    def tokenize(
        self,
        prompts: Union[list[str], list[PromptPattern]],
        return_tokens: bool = False,
        options: Options = None,
    ) -> list[TokenizeResult]:
        """The tokenize endpoint allows you to check the conversion of provided prompts to tokens
        for a given model. It splits text into words or subwords, which then are converted to ids
        through a look-up table (vocabulary). Tokenization allows the model to have a reasonable
        vocabulary size.

        Args:
            prompts (list[str]): The list of one or more prompt strings.
            return_tokens (bool, optional): Return tokens with the response. Defaults to False.
            options (Options, optional): Additional parameters to pass in the query payload. Defaults to None.

        Returns:
            list[TokenizeResult]: The Tokenized input
        """
        return list(self.tokenize_as_completed(prompts, return_tokens, options=options))

    def tokenize_async(
        self,
        prompts: Union[list[str], list[PromptPattern]],
        ordered: bool = False,
        callback: Callable[[TokenizeResult], Any] = None,
        return_tokens: bool = False,
        options: Options = None,
        *,
        throw_on_error: bool = False,
    ) -> Generator[Union[TokenizeResult, None]]:
        """The tokenize endpoint allows you to check the conversion of provided prompts to tokens
        for a given model. It splits text into words or subwords, which then are converted to ids
        through a look-up table (vocabulary). Tokenization allows the model to have a reasonable
        vocabulary size. This python method utilizes async capabilities and returns
        responses as they arrive.

        Args:
            prompts (list[str]): The list of one or more prompt strings.
            ordered (bool): Whether the responses should be returned in-order.
            callback (Callable[[TokenizeResult], Any]): Callback to call for each result.
            return_tokens (bool, optional): Return tokens with the response. Defaults to False.
            options (Options, optional): Additional parameters to pass in the query payload. Defaults to None.
            throw_on_error (bool, optional): Throws error on failure. Defaults to False.

        Returns:
            Generator[Union[TokenizeResult, None]]: The Tokenized input
        """
        if len(prompts) > 0 and isinstance(prompts[0], PromptPattern):
            prompts = PromptPattern.list_str(prompts)

        logger.debug(f"Calling Tokenize Async. Prompts: {prompts}")

        try:
            params = TokenParams(return_tokens=return_tokens)
            with AsyncResponseGenerator(
                self.model,
                prompts,
                params,
                self.service,
                fn="tokenize",
                ordered=ordered,
                callback=callback,
                options=options,
                throw_on_error=throw_on_error,
            ) as asynchelper:
                for response in asynchelper.generate_response():
                    yield response
        except Exception as ex:
            raise to_genai_error(ex)

    def tune(
        self,
        name: str,
        method: str,
        task: str,
        hyperparameters: CreateTuneHyperParams = None,
        training_file_ids: list[str] = None,
        validation_file_ids: list[str] = None,
    ):
        """Tune the base-model for given training data.

        Args:
            name (str): Label for this tuned model.
            method (str): The list of one or more prompt strings.
            task (str): Task ID, could be "classification", "summarization", or "generation"
            hyperparameters (CreateTuneHyperParams): Tuning hyperparameters
            training_file_ids (list[str]): IDs for files with training data
            validation_file_ids (list[str]): IDs for files with validation data

        Returns:
            Model: An instance of tuned model
        """
        if training_file_ids is None:
            raise GenAiException(ValueError("Parameter should be specified: training_file_paths or training_file_ids."))

        params = CreateTuneParams(
            name=name,
            model_id=self.model,
            method_id=method,
            task_id=task,
            training_file_ids=training_file_ids,
            validation_file_ids=validation_file_ids,
            parameters=hyperparameters or CreateTuneHyperParams(),
        )
        tune = TuneManager.create_tune(service=self.service, params=params)
        return Model(model=tune.id, params=None, credentials=self.creds)

    def status(self) -> str:
        """Get status of a tuned model.

        Returns:
            str: Status of a tuned model
        """
        tune = TuneManager.get_tune(tune_id=self.model, service=self.service)
        return tune.status

    def delete(self):
        params = TunesListParams()
        tunes = TuneManager.list_tunes(service=self.service, params=params).results
        id_to_status = {t.id: t.status for t in tunes}
        if self.model not in id_to_status:
            raise GenAiException(ValueError("Tuned model not found. Currently method supports only tuned models."))
        TuneManager.delete_tune(service=self.service, tune_id=self.model)

    def download(self):
        enconder_params = DownloadAssetsParams(id=self.model, content="encoder")
        logs_params = DownloadAssetsParams(id=self.model, content="logs")
        TuneManager.download_tune_assets(service=self.service, params=enconder_params)
        TuneManager.download_tune_assets(service=self.service, params=logs_params)

    @staticmethod
    def models(credentials: Credentials = None, service: ServiceInterface = None) -> list[ModelCard]:
        """Get a list of models

        Args:
            credentials (Credentials): Credentials
            service (ServiceInterface): Service Interface

        Returns:
            list[ModelCard]: A list of available models
        """
        service = _get_service(credentials, service)
        response = service.models()
        cards = ModelList(**response.json()).results
        return cards

    def available(self) -> bool:
        """Check if the model is available. Note that for tuned models
        the model could still be in the process of tuning.

        Returns:
            bool: Boolean indicating model availability
        """
        idset = set(m.id for m in Model.models(service=self.service))
        return self.model in idset

    def info(self) -> Union[ModelCard, None]:
        """Get info of the model

        Returns:
            Union[ModelCard, TuneInfoResult, None]: Model info
        """
        id_to_model = {m.id: m for m in Model.models(service=self.service)}
        return id_to_model.get(self.model, None)
